import os

import numpy as np

from alabamaEncode.conent_analysis.chunk.analyze_steps.new_grain import (
    calc_grainsynth_of_scene_fast,
)
from alabamaEncode.conent_analysis.chunk.analyze_steps.per_scene_grain import (
    calc_grainsynth_of_scene,
)
from alabamaEncode.scene.scene_detection import scene_detect


def gather_src_vals():
    test_env = "./experiments/grain_synth/"
    test_env = os.path.abspath(test_env)

    if not os.path.exists(test_env):
        os.makedirs(test_env)

    files = [
        "/home/kokoniara/ep3_halo_test.mkv",
        "/home/kokoniara/laPuczajna_fallout_2024_e1s01_testClip_ss840_t60.mkv",
        "/home/kokoniara/howlscastle_test.mkv",
    ]
    chunks = []
    for i, file in enumerate(files):
        scene_list = scene_detect(
            input_file=file,
            cache_file_path=f"{test_env}{i}_sceneCache.pt",
        )
        chunks.extend(scene_list.chunks)

    x = []
    y = []

    for chunk in chunks:
        src_values = calc_grainsynth_of_scene_fast(
            chunk,
            crop_vf="3840:1920:0:120",
            scale_vf="1920:-2",
            return_source_values=True,
        )
        try:
            grain_value = calc_grainsynth_of_scene(
                chunk,
                crop_vf="3840:1920:0:120",
                scale_vf="1920:-2",
                # parallel=True,
            )
        except RuntimeError:
            continue
        ref_size, weak_size, strong_size = (
            src_values["ref_size"],
            src_values["weak_size"],
            src_values["strong_size"],
        )
        # normalize using ref_size as 1
        weak_size /= ref_size
        strong_size /= ref_size
        print(f"weak: {weak_size}, strong: {strong_size}, grain: {grain_value}")
        x.append([weak_size, strong_size])
        y.append(grain_value)

    return x, y


if __name__ == "__main__":
    load = True
    if load:
        output_vals = np.load("output_vals.npy")
        training_data = np.load("training_data.npy")
    else:
        output_vals, training_data = gather_src_vals()
        output_vals = np.array(output_vals)

        # save
        np.save("output_vals.npy", output_vals)
        np.save("training_data.npy", training_data)

    print(f"Processed {len(training_data)} chunks")

    data = output_vals
    answers = training_data

    # output_vals = np.array(
    #     [
    #         [0.85019309, 0.83728725],
    #         [0.86150752, 0.8377991],
    #         [0.870051, 0.85019648],
    #         [0.87227345, 0.85443137],
    #         [0.86221075, 0.83221905],
    #         [0.87164853, 0.84443476],
    #         [0.87270245, 0.83833311],
    #         [0.84947409, 0.81849868],
    #         [0.86418557, 0.82854816],
    #     ]
    # )
    #
    # training_data = np.array([11, 9, 12, 8, 12, 14, 17, 14, 18])

    def polynomial_features(X, degree):
        from itertools import combinations_with_replacement

        n_samples, n_features = X.shape
        combinations = list(combinations_with_replacement(range(n_features), degree))
        X_poly = np.ones((n_samples, len(combinations)))

        for i, index_comb in enumerate(combinations):
            X_poly[:, i] = np.prod(X[:, index_comb], axis=1)

        return X_poly

    # Define the model function
    def predict(X, weights):
        return np.dot(X, weights)

    # Normal Equation method to find optimal weights
    def fit_normal_equation(X, y):
        X_b = np.c_[np.ones((X.shape[0], 1)), X]  # Add bias term
        return np.linalg.inv(X_b.T.dot(X_b)).dot(X_b.T).dot(y)

    # Create polynomial features
    degree = 2
    data_poly = polynomial_features(data, degree)

    weights = fit_normal_equation(data_poly, answers)

    def inference(new_data, weights, degree):
        new_data_poly = polynomial_features(new_data, degree)
        new_data_poly_b = np.c_[
            np.ones((new_data_poly.shape[0], 1)), new_data_poly
        ]  # Add bias term
        return predict(new_data_poly_b, weights)

    predictions = inference(data, weights, degree)
    errors = np.abs(predictions - answers)
    error_percentages = (errors / answers) * 100

    for real, pred, error, error_percent in zip(
        answers, predictions, errors, error_percentages
    ):
        print(
            f"real: {real}, predicted: {pred}, error: {error}, error percent: {error_percent}"
        )

    overall_error_percent = np.mean(error_percentages)
    print(f"overall error percent: {overall_error_percent}")

    # Example usage of the inference function
    new_data = np.array([[0.850193091421479, 0.8372872506855793]])
    new_predictions = inference(new_data, weights, degree)
    print("Predictions for new data:", int(new_predictions[0]))

    # print weights without e+x syntax
    print("Model weights")
    print(np.array2string(weights, separator=", "))



"""
big inspiration from:
https://github.com/porcino/Av1ador/blob/4b58a460000acdaee669b61a6e8500c925e3c3bd/Av1ador/Video.cs#L436
"""

import copy
import os
import subprocess
from math import sqrt
from statistics import mean

from alabamaEncode.conent_analysis.chunk.chunk_analyse_step import (
    ChunkAnalyzePipelineItem,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.timer import Timer
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.scene.chunk import ChunkObject
from alabamaEncode.scene.scene_detection import scene_detect


class GrainSynth(ChunkAnalyzePipelineItem):
    def run(self, ctx: AlabamaContext, chunk: ChunkObject, enc: Encoder) -> Encoder:
        grain_synth_result = ctx.get_kv().get("grain_synth", chunk.chunk_path)

        if grain_synth_result is None:
            grain_synth_result = calc_grainsynth_of_scene(
                chunk,
                scale_vf=ctx.scale_string,
                crop_vf=ctx.crop_string,
            )
            grain_synth_result = min(grain_synth_result, 18)

            enc.grain_synth = int(grain_synth_result)
            ctx.get_kv().set("grain_synth", chunk.chunk_path, grain_synth_result)
        else:
            enc.grain_synth = int(grain_synth_result)

        # if ctx.simple_denoise and grain_synth_result != 0:
        #     grain_synth_result /= 1.5
        ctx.log(f"{chunk.log_prefix()}computed gs {enc.grain_synth}", category="grain")

        return enc


def calc_grainsynth_of_scene(
    _chunk: ChunkObject,
    encoder_max_grain: int = 50,
    scale_vf="",
    crop_vf="",
    print_timing=False,
) -> int:
    chunk: ChunkObject = copy.deepcopy(_chunk)

    filter_vec = []
    if crop_vf != "" and crop_vf is not None:
        filter_vec.append(f"crop={crop_vf}")

    if scale_vf != "" and scale_vf is not None:
        filter_vec.append(f"scale={scale_vf}:flags=lanczos")

    denoise_weak = "nlmeans=s=1.8:p=7:r=9"
    denoise_strong = "nlmeans=s=4:p=5:r=15"

    timer = Timer()

    timer.start("scene_gs_approximation")

    filters_ref = ",".join(filter_vec)
    if filters_ref != "" and "vf" not in filters_ref:
        filters_ref = f" -vf {filters_ref}"
    filters_weak = ",".join(filter_vec + [denoise_weak])
    filters_weak = f" -vf {filters_weak}"
    filters_strong = ",".join(filter_vec + [denoise_strong])
    filters_strong = f" -vf {filters_strong}"

    gs = []

    chunk_frame_lenght = chunk.last_frame_index - chunk.first_frame_index
    loop_count = int((5 + sqrt(chunk_frame_lenght / 5)) / 2)  # magic
    for i in range(loop_count):
        curr = chunk.first_frame_index + int((chunk_frame_lenght / loop_count) * i)
        chunk.first_frame_index = curr

        if chunk.first_frame_index > chunk.last_frame_index:
            chunk.first_frame_index = chunk.last_frame_index - 1

        common = (
            f"{get_binary('ffmpeg')} -v error "
            f"-y {chunk.get_ss_ffmpeg_command_pair()} -frames:v 1 -pix_fmt yuv420p "
            f" -f image2pipe -c:v png "
        )

        timer.start(f"calc_scene_{i}")

        ref_size = get_size(f"{common} {filters_ref} -")
        if ref_size == 0:
            continue
        weak_size = get_size(f"{common} {filters_weak} -")
        if weak_size == 0:
            continue
        strong_size = get_size(f"{common} {filters_strong} -")
        if strong_size == 0:
            continue

        timer.stop(f"calc_scene_{i}")

        grain_factor = ref_size * 100.0 / weak_size
        grain_factor = (
            ((ref_size * 100.0 / strong_size * 100.0 / grain_factor) - 105.0)
            * 8.0
            / 10.0
        )

        # print(f"Frame grain: {int(grain_factor / (100.0 / encoder_max_grain))}")

        # magic empirical values
        grain_factor = max(0.0, min(100.0, grain_factor))
        gs.append(grain_factor)

    if len(gs) == 0:
        raise RuntimeError("No grain factors found")

    final_grain = mean(gs)
    final_grain /= 100.0 / encoder_max_grain
    final_grain = int(final_grain)

    timer.stop("scene_gs_approximation")
    timer.finish(loud=print_timing)

    return final_grain


def get_size(command):
    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    total_size = 0

    while process.poll() is None:
        chunk = process.stdout.read(1024 * 1024)
        if chunk:
            total_size += len(chunk)

    process.wait()

    return total_size


def test():
    test_env = "./experiments/grain_synth/"
    test_env = os.path.abspath(test_env)

    if not os.path.exists(test_env):
        os.makedirs(test_env)

    scene_list = scene_detect(
        input_file="/home/kokoniara/ep3_halo_test.mkv",
        cache_file_path=test_env + "sceneCache.pt",
    )

    grain_values = {}
    for chunk in scene_list.chunks:
        grain_value = calc_grainsynth_of_scene(
            chunk,
            crop_vf="3840:1920:0:120",
            scale_vf="1920:-2",
            # parallel=True,
        )
        print("calculated gs: " + str(grain_value))
        grain_values[str(chunk.chunk_index)] = grain_value

        print("\n\n")

    print(grain_values)


if __name__ == "__main__":
    test()



import os

from alabamaEncode.core.util.bin_utils import get_binary, verify_ffmpeg_library
from alabamaEncode.core.util.cli_executor import run_cli


class AvifEncoderSvtenc:
    """
    AVIF Encoder but SvtAv1 inside ffmpeg.
    """

    vf = ""
    grain_synth = 0
    speed = 3
    bit_depth = 10
    crf = 13
    passes = 1
    bitrate = 1000
    output_path = ""
    in_path = ""

    def get_encode_commands(self) -> str:
        verify_ffmpeg_library("libsvtav1")
        if not self.output_path.endswith(".avif"):
            raise Exception("FATAL: output_path must end with .avif")

        if self.bit_depth not in (8, 10):
            raise Exception("FATAL: bit must be 8 or 10")

        pix_fmt = "yuv420p" if self.bit_depth == 8 else "yuv420p10le"

        if self.bitrate is not None and self.bitrate != -1:
            rate = f"-b:v {self.bitrate}k"
        else:
            rate = f"-crf {self.crf}"

        return (
            f'{get_binary("ffmpeg")} -hide_banner -y -i "{self.in_path}" {self.vf} '
            f"-c:v libsvtav1 {rate} "
            f"-svtav1-params tune=0:film-grain={self.grain_synth}"
            f' -preset {self.speed} -pix_fmt {pix_fmt} "{self.output_path}"'
        )

    def run(self):
        out = run_cli(self.get_encode_commands()).get_output()
        if (
            not os.path.exists(self.output_path)
            or os.path.getsize(self.output_path) < 1
        ):
            print(self.get_encode_commands())
            raise Exception(
                f"FATAL: SVTENC ({self.get_encode_commands()}) FAILED with " + out
            )



from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.impl.Aomenc import EncoderAom
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.encoder.impl.VideoToolbox import EncoderAppleHEVC


def convexhull_get_crf_range(codec: Codec) -> tuple[int, int]:
    match codec:
        case Codec.av1:
            return 18, 58
        # TODO: pick other values
        case Codec.h264:
            return 10, 40
        case Codec.h265:
            return 0, 51
        case Codec.vp8:
            return 0, 63
        case Codec.vp9:
            return 0, 63
        case _:
            raise ValueError(f"FATAL: Unknown codec: {codec}")


def get_vmaf_probe_speed(encoder: Encoder, ctx=None) -> int:
    if ctx is not None:
        if ctx.vmaf_probe_speed != -1:
            return ctx.vmaf_probe_speed
    match encoder:
        case EncoderSvt():
            return 13
        case _:
            # TODO: pick other values
            return 5


def get_vmaf_probe_offset(enc: Encoder) -> int:
    match enc:
        case EncoderSvt():
            return 1
        case _:
            # TODO: pick other values
            return 0


def convexhull_get_resolutions(codec: Codec) -> list[str]:
    match codec:
        case Codec.av1:
            return ["1920:-2", "1280:-2", "960:-2", "854:-2"]
        case _:
            return ["1920:-2", "1280:-2", "960:-2", "854:-2", "768:-2", "480:-2"]


def get_crf_limits(encoder: Encoder, ctx=None) -> tuple[int, int]:
    if ctx is not None:
        if ctx.crf_limits is not None:
            return tuple(map(int, ctx.crf_limits.split(",")))
    match encoder.get_codec():
        case Codec.av1:
            match encoder:
                case EncoderSvt():
                    return 20, 35
                case _:
                    return 18, 40
        case _:
            return 12, 50


if __name__ == "__main__":
    my_enc = EncoderSvt()
    my_aom_enc = EncoderAom()

    assert get_crf_limits(my_enc), [22, 38]
    assert get_crf_limits(my_aom_enc), [18, 40]
    assert get_crf_limits(EncoderAppleHEVC()), [10, 90]


def get_vmaf_list(codec: Codec) -> list[int]:
    match codec:
        case Codec.av1:
            return [73, 78, 84, 91, 95]
        case Codec.h264:
            return [45, 55, 62, 68, 81, 87, 90, 93, 95, 96]
        case Codec.vp9:
            return [73, 78, 84, 91, 95]
        case _:
            return [73, 78, 84, 91, 95]



import os

from alabamaEncode.core.util.bin_utils import get_binary, register_bin
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt


def build_svt():
    build_dir = "~/.alabamaEncoder/build"
    bin_dir = "~/.alabamaEncoder/bin"
    git_url = "https://gitlab.com/AOMediaCodec/SVT-AV1.git"

    build_dir = os.path.expanduser(build_dir)
    bin_dir = os.path.expanduser(bin_dir)

    bin_name = "SvtAv1EncApp"
    bin_path = bin_dir + "/" + bin_name

    if not os.path.exists(build_dir):
        os.makedirs(build_dir)

    if not os.path.exists(bin_dir):
        os.makedirs(bin_dir)

    repo_path = build_dir + "/SVT-AV1"

    if not os.path.exists(repo_path):
        print("Cloning SVT repo")
        os.system(
            f"{get_binary('git')} clone --depth 1 --single-branch --branch master --no-tags {git_url} \"{repo_path}\""
        )
    else:
        print("Updaing SVT repo")
        os.system(f"{get_binary('git')} pull \"{repo_path}\"")

    if not os.path.exists(repo_path):
        print("Failed to clone SVT repo")
        return

    print("Building SVT")
    os.system(
        f"cd {repo_path}/Build/linux && ./build.sh clean &&"
        f" ./build.sh cc=clang cxx=clang++ enable-lto static native release"
    )

    new_bin_path = repo_path + "/Bin/Release/SvtAv1EncApp"

    if os.path.exists(new_bin_path):
        os.system(f"chmod +x {new_bin_path}")
        register_bin("SvtAv1EncApp", new_bin_path)
        print(f"Just compiled SVT: {EncoderSvt().get_version()}")
        os.rename(new_bin_path, bin_path)
    else:
        print("Failed to compile SVT")


def build_vmaf():
    # build is a strong word...
    bin_url = "https://github.com/Netflix/vmaf/releases/download/v3.0.0/vmaf"

    bin_dir = os.path.expanduser("~/.alabamaEncoder/bin")
    bin_name = "vmaf"
    bin_path = bin_dir + "/" + bin_name

    if not os.path.exists(bin_dir):
        os.makedirs(bin_dir)
    import requests

    print("Downloading VMAF")

    r = requests.get(bin_url, allow_redirects=True)
    open(bin_path, "wb").write(r.content)
    if not os.path.exists(bin_path):
        print("Failed to download VMAF")
        return
    os.system(f"chmod +x {bin_path}")
    version = run_cli(f"{bin_path} -v").get_output()
    print(f"Just downloaded VMAF: {version}")


if __name__ == "__main__":
    # build_svt()
    build_vmaf()



import re
from copy import copy
from typing import List

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution


class EncoderAom(Encoder):
    def get_pretty_name(self) -> str:
        return "AOMENC"

    def get_codec(self) -> Codec:
        return Codec.av1

    def supports_grain_synth(self) -> bool:
        return True

    def get_version(self) -> str:
        # Included encoders:
        #
        # av1    - AOMedia Project AV1 Encoder Psy v3.6.0 (default)
        match = re.search(
            r"av1\s+-\s+AOMedia Project AV1 Encoder\s+(.*)",
            run_cli(f"{get_binary('aomenc')} --help").get_output(),
        )
        if match is None:
            raise Exception("FATAL: Could not find av1 encoder version")
        return match.group(1)

    def __init__(self):
        super().__init__()
        self.photon_noise_path = ""
        self.use_webm = False

    def get_encode_commands(self) -> List[str]:
        self.speed = min(self.speed, 9)
        encode_command = self.chunk.create_chunk_ffmpeg_pipe_command(
            video_filters=self.video_filters
        )
        encode_command += " | "
        encode_command += f"{get_binary('aomenc')} - "
        encode_command += " --quiet "
        encode_command += f'-o "{self.output_path}" '

        # 1 thread cuz we generally run this chunked
        encode_command += f"--threads={self.threads} "

        encode_command += f"--cpu-used={self.speed} "
        encode_command += f"--bit-depth={self.bit_override} "

        if self.override_flags == "" or self.override_flags is None:
            # STOLEN FROM ROOTATKAI IN #BENCHMARKS CUZ HE SAID ITS GOOD OR WHATEVER --lag-in-frames=48
            # --tune-content=psy --tune=ssim --sb-size=dynamic --enable-qm=1 --qm-min=0 --qm-max=8 --row-mt=1
            # --disable-kf --kf-max-dist=9999 --kf-min-dist=1 --disable-trellis-quant=0 --arnr-maxframes=15
            encode_command += f"--lag-in-frames=48 "
            encode_command += f"--tune-content=psy "
            encode_command += " --tune=ssim "
            encode_command += f"--sb-size=dynamic "
            encode_command += f"--enable-qm=1 "
            encode_command += f"--qm-min=0 "
            encode_command += f"--qm-max=8 "
            encode_command += f"--row-mt=1 "
            encode_command += f"--disable-kf "
            encode_command += f"--kf-max-dist=9999 "
            encode_command += f"--kf-min-dist=1 "
            encode_command += f"--disable-trellis-quant=0 "
            encode_command += f"--arnr-maxframes=15 "
        else:
            encode_command += self.override_flags

        if self.use_webm:
            encode_command += " --webm"
        else:
            encode_command += " --ivf"

        match self.rate_distribution:
            case EncoderRateDistribution.VBR:
                encode_command += f" --end-usage=vbr --target-bitrate={self.bitrate}"
            case EncoderRateDistribution.CQ:
                encode_command += f" --end-usage=q --cq-level={self.crf}"
            case EncoderRateDistribution.CQ_VBV:
                if self.bitrate != -1 and self.bitrate is not None:
                    encode_command += f" --end-usage=cq --cq-level={self.crf} --target-bitrate={self.bitrate} "
                else:
                    encode_command += f" --end-usage=q --cq-level={self.crf}"

        if self.photon_noise_path == "":
            encode_command += (
                f" --enable-dnl-denoising=0 --denoise-noise-level={self.grain_synth}"
            )
        else:
            encode_command += f' --enable-dnl-denoising=0 --film-grain-table="{self.photon_noise_path}"'

        if self.passes == 2:
            encode_command += f' --fpf="{self.output_path}.log"'
            encode_command += " --passes=2"

            pass2 = copy(encode_command)
            pass2 += " --pass=2"

            pass1 = copy(encode_command)
            pass1 += " --pass=1"

            return [pass1, pass2, f"rm {self.output_path}.log"]
        else:
            encode_command += " --pass=1"
            return [encode_command]

    def get_chunk_file_extension(self) -> str:
        return ".ivf"



import os

from alabamaEncode.conent_analysis.chunk.final_encode_step import (
    FinalEncodeStep,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.calculate import get_metric_from_stats
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.chunk import ChunkObject


class DynamicTargetVmafVBR(FinalEncodeStep):
    def run(
        self, enc: Encoder, chunk: ChunkObject, ctx: AlabamaContext, encoded_a_frame
    ) -> EncodeStats:
        vmaf_options = ctx.get_vmaf_options()
        probe_file_base = ctx.get_probe_file_base(chunk.chunk_path)
        enc.passes = 3
        enc.rate_distribution = EncoderRateDistribution.VBR
        metric_name = "vmaf"
        original_output_path = enc.output_path
        low_bitrate = 250
        low_bitrate_cap = 250
        high_bitrate = 5000
        high_bitrate_cap = 20_000
        target_vmaf = ctx.vmaf
        trys = []
        max_vmaf_error = 0.5
        stats = None
        binary_search_steps = 4
        interpol_steps = 3
        mid_bitrate = 2000

        total_encodes = 0

        def log(_str):
            ctx.log(
                f"{chunk.log_prefix()}{_str}",
                category="probe",
            )

        def get_score(_bitrate):
            enc.bitrate = _bitrate
            enc.output_path = os.path.join(
                probe_file_base,
                f"{chunk.chunk_index}.{_bitrate}{enc.get_chunk_file_extension()}",
            )
            stats: EncodeStats = enc.run(
                metric_to_calculate=Metric.VMAF,
                metric_params=vmaf_options,
                override_if_exists=False,
            )
            nonlocal total_encodes
            total_encodes += 1
            metric = get_metric_from_stats(
                stats, statistical_representation=ctx.vmaf_target_representation
            )
            metric_error = abs(metric - target_vmaf)
            log(
                f"bitrate: {stats.bitrate}; {metric_name}: {metric}; error: {metric_error}"
                f" attempt {total_encodes}/{binary_search_steps + interpol_steps}"
            )
            nonlocal trys
            trys.append((_bitrate, metric))
            return metric, enc.output_path, stats

        def finish(_stats, bitrate, additional_message=""):
            prefi = f"{additional_message}; " if additional_message != "" else ""
            log(
                f"{prefi}bitrate {bitrate}; {metric_name}: {get_metric_from_stats(_stats, ctx.vmaf_target_representation)};"
                f" real bitrate: {_stats.bitrate} kb/s"
            )
            ctx.get_kv().set("best_bitrates", chunk.chunk_index, bitrate)
            os.rename(enc.output_path, original_output_path)
            return _stats

        for search_step in range(binary_search_steps):
            if search_step > 0:
                mid_bitrate = (low_bitrate + high_bitrate) // 2
            elif search_step == 0:
                if len(ctx.best_crfs) > 2:
                    # set the mean of the best crf's so far as the first guess
                    best_bitrates = ctx.get_kv().get_all("best_bitrates")
                    mid_bitrate = int(sum(best_bitrates.values()) / len(best_bitrates))

                    log(f"overriding first attempt to {mid_bitrate}")

            mid_vmaf, probe_path, stats = get_score(mid_bitrate)

            current_metric_error = abs(target_vmaf - mid_vmaf)
            if current_metric_error < max_vmaf_error:
                return finish(
                    stats,
                    mid_bitrate,
                    additional_message=f"Probe within error ({current_metric_error})",
                )

            # check if two probes had the same vmaf, if so quit early since its prob a black screen
            if len(trys) > 1 and trys[-1][1] == trys[-2][1]:
                return finish(
                    stats,
                    mid_bitrate,
                    additional_message=f"Two probes same {metric_name} quiting",
                )

            if mid_vmaf < target_vmaf:
                low_bitrate = mid_bitrate
            else:
                high_bitrate = mid_bitrate

        for interpol_step in range(interpol_steps):
            # sort by vmaf difference from target
            points = sorted(trys, key=lambda _x: abs(_x[1] - target_vmaf))

            # get the two closest points
            bitrate_low, metric_low = points[0]
            bitrate_high, metric_high = points[1]

            # if the difference bettwen metric_low and metric_high is less than metric_high and metric_high,
            # then use the high point instead of the low point
            if len(points) > 2:
                if abs(metric_low - metric_high) < abs(metric_high - points[2][1]):
                    bitrate_high, metric_high = points[2]

            interpolated_bitrate = bitrate_low + (bitrate_high - bitrate_low) * (
                (target_vmaf - metric_low) / (metric_high - metric_low)
            )

            interpolated_bitrate = max(
                min(interpolated_bitrate, high_bitrate_cap), low_bitrate_cap
            )

            interpolated_bitrate = round(interpolated_bitrate)

            # quit early if interpolated bitrate is in trys
            if any([abs(_x[0] - interpolated_bitrate) < 25 for _x in trys]):
                return finish(
                    stats,
                    interpolated_bitrate,
                    additional_message=f"intrpol bitrate {interpolated_bitrate} already tried",
                )

            log(f"Interpolate pass {interpol_step}: bitrate {interpolated_bitrate}")
            metric, probe_path, stats = get_score(interpolated_bitrate)
            current_metric_error = abs(target_vmaf - metric)
            if current_metric_error <= max_vmaf_error:
                return finish(
                    stats,
                    interpolated_bitrate,
                    additional_message=f"Probe within error ({current_metric_error})",
                )

        return finish(
            stats, interpolated_bitrate, additional_message="Failed to hit target"
        )

    def dry_run(self, enc: Encoder, chunk: ChunkObject) -> str:
        raise Exception(f"dry_run not implemented for {self.__class__.__name__}")



import re
from typing import List

from alabamaEncode.core.util.bin_utils import get_binary, check_bin
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution


class EncoderSvt(Encoder):
    def get_pretty_name(self) -> str:
        return "SVT_AV1"

    def get_codec(self) -> Codec:
        return Codec.av1

    def supports_grain_synth(self) -> bool:
        return True

    def get_encode_commands(self) -> List[str]:
        if (
            self.keyint == -1 or self.keyint == -2
        ) and self.rate_distribution == EncoderRateDistribution.VBR:
            print("WARNING: keyint must be set for VBR, setting to 240")
            self.keyint = 240

        kommand = ""

        if check_bin("taskset"):
            if self.pin_to_core != -1:
                kommand += f"taskset -a -c {self.pin_to_core} "

        kommand += (
            f"{self.get_ffmpeg_pipe_command()} | "
            f"{get_binary('SvtAv1EncApp')}"
            f" -i stdin"
            f" --input-depth {self.bit_override}"
            f" --progress 2 "
        )

        match self.rate_distribution:
            case EncoderRateDistribution.CQ:
                if self.passes != 1:
                    print("WARNING: passes must be 1 for CQ, setting to 1")
                    self.passes = 1
                kommand += f" --crf {self.crf} --rc 0"
            case EncoderRateDistribution.VBR:
                kommand += f" --rc 1 --tbr {self.bitrate} --undershoot-pct 95 --overshoot-pct 10 "
            case EncoderRateDistribution.CQ_VBV:
                kommand += f" --crf {self.crf} --mbr {self.bitrate}"
            case EncoderRateDistribution.VBR_VBV:
                raise Exception("FATAL: VBR_VBV is not supported")

        kommand += f" --pin 0"
        kommand += f" --lp {self.threads}"

        kommand += f" --preset {self.speed}"

        # check one be one if any flags are in the override flags, if not use the framework ones
        def add_flag(flag, value):
            nonlocal kommand
            if flag not in self.override_flags:
                kommand += f" {flag} {value}"

        if 0 <= self.grain_synth <= 50 and "--film-grain":
            add_flag("--film-grain", self.grain_synth)

        add_flag("--color-primaries", self.color_primaries)
        add_flag("--transfer-characteristics", self.transfer_characteristics)
        add_flag("--matrix-coefficients", self.matrix_coefficients)

        if self.hdr:
            add_flag("--enable-hdr", 1)
            add_flag("--chroma-sample-position", self.chroma_sample_position)
            add_flag(
                "--content-light",
                f"{self.maximum_content_light_level},{self.maximum_frame_average_light_level}",
            )

            if self.svt_master_display != "":
                add_flag("--mastering-display", f'"{self.svt_master_display}"')

        add_flag("--tune", self.svt_tune)
        add_flag("--aq-mode", self.svt_aq_mode)
        add_flag("--tile-columns", self.tile_cols)
        add_flag("--tile-rows", self.tile_rows)
        add_flag("--keyint", self.keyint)

        if self.qm_enabled:
            add_flag("--enable-qm", 1)
            add_flag("--qm-min", self.qm_min)
            add_flag("--qm-max", self.qm_max)
        else:
            add_flag("--enable-qm", 0)

        add_flag("--film-grain-denoise", 0)
        add_flag("--enable-tf", self.svt_tf)
        add_flag("--enable-variance-boost", self.svt_enable_variance_boost)
        add_flag("--variance-boost-strength", self.svt_variance_boost_strength)
        add_flag("--variance-octile", self.svt_variance_octile)

        if self.is_psy():
            add_flag("--sharpness", self.svt_sharpness)

        if self.svt_supperres_mode != 0:
            add_flag("--superres-mode", self.svt_supperres_mode)
            add_flag("--superres-denom", self.svt_superres_denom)
            add_flag("--superres-kf-denom", self.svt_superres_kf_denom)
            add_flag("--superres-qthres", self.svt_superres_qthresh)
            add_flag("--superres-kf-qthres", self.svt_superres_kf_qthresh)

        if self.svt_sframe_interval > 0:
            add_flag("--sframe-dist", self.svt_sframe_interval)
            add_flag("--sframe-mode", self.svt_sframe_mode)

        if self.svt_resize_mode != 0:
            add_flag("--resize-mode", self.svt_resize_mode)
            add_flag("--resize-denominator", self.svt_resize_denominator)
            add_flag("--resize-kf-denominator", self.svt_resize_kf_denominator)

        if self.override_flags != "":
            kommand += " " + self.override_flags + " "

        stats_bit = ""

        if self.passes > 1:
            stats_bit = f"--stats {self.output_path}.stat"

        match self.passes:
            case 2:
                commands = [
                    f"{kommand} --pass 1 {stats_bit}",
                    f"{kommand} --pass 2 {stats_bit} -b {self.output_path}",
                    f"rm {self.output_path}.stat",
                ]
            case 1:
                commands = [f'{kommand} -b "{self.output_path}"']
            case 3:
                commands = [
                    f"{kommand} --pass 1 {stats_bit}",
                    f"{kommand} --pass 2 {stats_bit}",
                    f"{kommand} --pass 3 {stats_bit} -b {self.output_path}",
                    f"rm {self.output_path}.stat",
                ]
            case _:
                raise Exception(f"FATAL: invalid passes count {self.passes}")

        return commands

    def get_chunk_file_extension(self) -> str:
        return ".ivf"

    def get_version(self) -> str:
        # Svt[info]: -------------------------------------------
        # Svt[info]: SVT [version]:	SVT-AV1 Encoder Lib v1.7.0-2-g09df835
        o = run_cli(f"{get_binary('SvtAv1EncApp')} --version").get_output()
        return " ".join(o.split(" ")[:-1])

    def is_psy(self) -> bool:
        return "PSY" in self.get_version()

    def parse_output_for_output(self, buffer) -> List[str]:
        if buffer is None:
            return []
        match = re.search(r"Encoding frame .+\d f", buffer)
        if match:  # check if we are past the header, also extract the string
            _match = re.search(
                r"Encoding\sframe\s+([0-9]+)\s([0-9.]+)\s.+\s([0-9.]+)\sf",
                match.group(0),
            )  # parse out the frame number, time, and bitrate
            return [_match.group(1), _match.group(2), _match.group(3)]
        else:
            return []



import asyncio
import os
import time
from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple

import psutil
from tqdm import tqdm

from alabamaEncode.core.chunk_job import ChunkEncoder
from alabamaEncode.parallel_execution.celery_app import run_command_on_celery, app
from alabamaEncode.parallel_execution.command import BaseCommandObject


async def execute_commands(
    use_celery=False,
    command_objects: List[BaseCommandObject] = None,
    multiprocess_workers: int = -1,
    pin_to_cores=False,
    finished_scene_callback: callable = None,
    size_estimate_data: tuple = None,
    throughput_scaling=False,
    pbar: tqdm = None,
):
    """
    Execute a list of commands in parallel
    :param throughput_scaling:
    :param pbar:
    :param pin_to_cores:
    :param use_celery: execute on a celery cluster
    :param command_objects: objects with a `run()` method to execute
    :param multiprocess_workers: number of workers in multiprocess mode, -1 for auto adjust
    :param finished_scene_callback: call when a scene finishes, contains the number of finished scenes
    :param size_estimate_data: tuple(frames, kB) of scenes encoded so far for the estimate
    """
    if command_objects is None or len(command_objects) == 0:
        return
    are_commands_adaptive_commands = isinstance(command_objects[0], ChunkEncoder)

    # this is for pure ide convince; "typecast" command objects
    if are_commands_adaptive_commands:
        command_objects: List[ChunkEncoder] = command_objects

    # values for a bitrate estimate in the progress bar
    encoded_frames_so_far = 0
    encoded_size_so_far = 0  # in kbps, kilo-bits per second
    if size_estimate_data is not None:
        encoded_frames_so_far, encoded_size_so_far = size_estimate_data

    total_scenes = len(command_objects)

    if use_celery:
        for a in command_objects:
            a.run_on_celery = True

        running_tasks = [
            run_command_on_celery.delay(command) for command in command_objects
        ]
        pbar.set_description(f"WORKERS - ESTM BITRATE -")

        while True:
            try:
                num_workers = len(app.control.inspect().active_queues())
                if num_workers is None or num_workers < 0:
                    print(
                        "No workers available, waiting for workers to become available"
                    )
                if all(task.ready() for task in running_tasks):
                    break
                for task in running_tasks:
                    if task.ready():
                        result = task.get()
                        if are_commands_adaptive_commands and result is not None:
                            finished_command = command_objects[
                                running_tasks.index(task)
                            ]
                            pbar.update(finished_command.chunk.get_frame_count())
                            pinned_code, stats = result
                            encoded_frames_so_far += stats["length_frames"]
                            encoded_size_so_far += stats["size"]
                            bitrate_estimate = "ESTM BITRATE -"
                            if encoded_frames_so_far > 0:
                                fps = command_objects[0].chunk.framerate
                                bitrate_estimate = (
                                    f"ESTM BITRATE {((encoded_size_so_far * 8) / (encoded_frames_so_far / fps)):.2f} "
                                    f"kb/s"
                                )

                            pbar.set_description(
                                f"WORKERS {num_workers} {bitrate_estimate}"
                            )
                        else:
                            pbar.update()
                        running_tasks.remove(task)
                await asyncio.sleep(1)
            except KeyboardInterrupt as e:
                print("Keyboard interrupt, cancelling tasks")
                for task in running_tasks:
                    task.revoke()
                raise e
        pbar.close()
    else:
        futures, completed_count = [], 0

        core_count, auto_scale = os.cpu_count(), multiprocess_workers == -1
        target_cpu_utilization, max_mem_usage = 95, 80

        max_jobs_limit = (
            core_count * 2
            if auto_scale
            else multiprocess_workers
            if multiprocess_workers != -1
            else core_count
        )

        local_jobs_limit = multiprocess_workers if not auto_scale else 2

        if throughput_scaling:
            local_jobs_limit = core_count / 2

        loop, executor = asyncio.get_event_loop(), ThreadPoolExecutor()

        # array of zeros to keep track of which cores are used, used for thread pinning with taskset
        used_cores = [0] * core_count

        frame_encoded_history: List[Tuple[int, float]] = []
        last_frame_encode = time.time()
        thughput_history = []

        current_thughput = -1
        previuus_thughput = -1
        thouput_compare = -1

        # -1 for decreasing, 0 for stable, 1 for increasing
        thouput_change_trend = 0
        thouput_reverse_trend_trigger_counter = 0

        def frames_encoded(frame_count=1):
            nonlocal last_frame_encode
            nonlocal frame_encoded_history
            nonlocal current_thughput
            nonlocal previuus_thughput
            nonlocal thughput_history
            now = time.time()
            if last_frame_encode == -1:
                last_frame_encode = now
            time_since_update = now - last_frame_encode
            last_frame_encode = now
            frame_encoded_history.append((frame_count, time_since_update))

            # measure the throughput based on the last 5 updates, save to history and print
            if len(frame_encoded_history) > 5:
                frame_encoded_history.pop(0)
            total_frames = sum([f[0] for f in frame_encoded_history])
            total_time = sum([f[1] for f in frame_encoded_history])
            thughput = total_frames / total_time
            thughput_history.append(thughput)
            # calculate the average throughput from the last 3 updates
            if len(thughput_history) > 3:
                thughput_history.pop(0)
            previuus_thughput = current_thughput
            current_thughput = sum(thughput_history) / len(thughput_history)
            # tqdm.write(
            #     f"Current throughput: {current_thughput:.2f} f/s, Last: {previuus_thughput:.2f} f/s, currently running jobs: {len(futures)}"
            # )

        def callback_wrapper():
            pbar.update()
            frames_encoded()

        # in the case that the encoder class supports "encoded a frame callback",
        # we can update the progress bar immediately after a frame is encoded,
        # so we set the callback
        if are_commands_adaptive_commands:
            for c in command_objects:
                c.encoded_a_frame_callback = (
                    lambda frame, bitrate, fps: callback_wrapper()
                )

        pbar.set_description(f"WORKERS: - CPU -% SWAP -%")

        while completed_count < total_scenes:
            currently_running_jobs = len(futures)

            # Start new jobs if we are under the local_jobs_limit
            while (
                currently_running_jobs <= local_jobs_limit
                and completed_count + currently_running_jobs < total_scenes
            ):
                command = command_objects[completed_count + currently_running_jobs]
                if pin_to_cores and 0 in used_cores:
                    core = used_cores.index(0)
                    used_cores[core] = 1
                    command.pin_to_core = core
                future = loop.run_in_executor(executor, command.run)
                futures.append(future)
                currently_running_jobs = len(futures)

            # wait for any of the tasks to finish
            done, pending = await asyncio.wait(futures, timeout=7)
            futures = [f for f in futures if not f.done()]
            currently_running_jobs = len(futures)

            # do stuff with the finished tasks
            for future in done:
                rslt = await future
                units_encoded = 1
                if are_commands_adaptive_commands and rslt is not None:
                    pined_core: int = rslt[0]
                    stats = rslt[1]
                    command_object = command_objects[completed_count]
                    if not command_object.supports_encoded_a_frame_callback():
                        units_encoded = command_object.chunk.get_frame_count()
                        frames_encoded(units_encoded)
                    if stats is not None:
                        encoded_frames_so_far += stats["length_frames"]
                        encoded_size_so_far += stats["size"]

                    if pin_to_cores:
                        used_cores[pined_core] = 0

                pbar.update(units_encoded)

                completed_count += 1

            # update the progress bar with the current system stats
            bitrate_estimate = " ESTM BITRATE N/A"
            if encoded_frames_so_far > 0:
                fps = command_objects[0].chunk.framerate
                bitrate_estimate = (
                    f" ESTM BITRATE {((encoded_size_so_far * 8) / (encoded_frames_so_far / fps)):.2f} "
                    f"kb/s"
                )
            cpu_percent = psutil.cpu_percent()
            memory_percent = psutil.virtual_memory().percent
            pbar.set_description(
                f"WORKERS {currently_running_jobs} CPU {int(cpu_percent)}% "
                f"MEM {int(memory_percent)}%{bitrate_estimate}"
            )
            # change the local_jobs_limit based on the picked strategy
            if throughput_scaling:
                if thouput_compare != current_thughput:
                    tqdm.write(
                        f"Checking throughput, current: {current_thughput:.2f} f/s, previous: {previuus_thughput:.2f} f/s"
                    )
                    # if we observe that throughput is decreasing, reverse the trend
                    if previuus_thughput > current_thughput:
                        thouput_reverse_trend_trigger_counter += 1
                        if thouput_reverse_trend_trigger_counter <= 2:
                            thouput_reverse_trend_trigger_counter = 0
                            if thouput_change_trend == -1:
                                thouput_change_trend = 1
                            elif thouput_change_trend == 1:
                                thouput_change_trend = -1
                            tqdm.write(
                                f"Reversing jobs limit trend to {'increasing' if thouput_change_trend == 1 else 'decreasing'}"
                            )
                            if thouput_change_trend == 1:
                                tqdm.write(
                                    f"Increasing local_jobs_limit to {local_jobs_limit + 1}"
                                )
                                local_jobs_limit += 1
                            # if the throughput is decreasing, decrease the local_jobs_limit
                            elif thouput_change_trend == -1:
                                tqdm.write(
                                    f"Decreasing local_jobs_limit to {local_jobs_limit - 1}"
                                )
                                local_jobs_limit -= 1

                            thouput_compare = current_thughput
            elif auto_scale and currently_running_jobs > 0:             
                local_jobs_limit += (
                    1
                    if cpu_percent <= target_cpu_utilization
                    and memory_percent <= max_mem_usage
                    else -1
                )
                local_jobs_limit = max(1, min(local_jobs_limit, max_jobs_limit))

            if finished_scene_callback is not None:
                finished_scene_callback(completed_count)

        pbar.close()



from typing import List

from alabamaEncode.conent_analysis.chunk.analyze_steps.multires_encode_candidates import (
    EncodeMultiResCandidates,
)
from alabamaEncode.conent_analysis.chunk.final_encode_steps.dynamic_target_vmaf import (
    DynamicTargetVmaf,
)
from alabamaEncode.conent_analysis.chunk.final_encode_steps.multires_encode_finals import (
    EncodeMultiResFinals,
)
from alabamaEncode.conent_analysis.refine_step import RefineStep
from alabamaEncode.conent_analysis.refine_steps.multires_package import MutliResPackage
from alabamaEncode.conent_analysis.refine_steps.multires_trellis import MutliResTrellis


def setup_chunk_analyze_chain(ctx, sequence):
    """
    Sets up the chunk analyze chain
    """
    from alabamaEncode.conent_analysis.chunk.analyze_steps.manual_crf import (
        CrfIndexesMap,
    )
    from alabamaEncode.conent_analysis.chunk.analyze_steps.optimised_vbr import (
        VbrPerChunkOptimised,
    )
    from alabamaEncode.conent_analysis.chunk.analyze_steps.per_scene_grain import (
        GrainSynth,
    )
    from alabamaEncode.conent_analysis.chunk.analyze_steps.new_grain import (
        NewGrainSynth,
    )
    from alabamaEncode.conent_analysis.chunk.analyze_steps.plain_crf import PlainCrf
    from alabamaEncode.conent_analysis.chunk.analyze_steps.plain_vbr import PlainVbr
    from alabamaEncode.conent_analysis.chunk.analyze_steps.target_vmaf import TargetVmaf

    ctx.chunk_analyze_chain = []

    # grain synth, per chunk parameters etc.
    if ctx.prototype_encoder.grain_synth == -2:
        ctx.chunk_analyze_chain.append(GrainSynth())
    elif ctx.prototype_encoder.grain_synth == -3:
        ctx.chunk_analyze_chain.append(NewGrainSynth())

    # setting up rate control, adaptive or not
    if ctx.bitrate_adjust_mode == "chunk":
        ctx.chunk_analyze_chain.append(VbrPerChunkOptimised())
    elif ctx.dynamic_vmaf_target:
        print("Using dynamic vmaf targeting")
        ctx.chunk_analyze_chain.append(PlainCrf())
    elif ctx.dynamic_vmaf_target_vbr:
        print("Using dynamic vmaf targeting with vbr")
        ctx.chunk_analyze_chain.append(PlainVbr())
    elif ctx.multi_res_pipeline:
        ctx.chunk_analyze_chain.append(EncodeMultiResCandidates())
    elif ctx.crf_map != "":
        ctx.chunk_analyze_chain.append(CrfIndexesMap(ctx.crf_map))
    elif ctx.crf_based_vmaf_targeting is True and ctx.multi_res_pipeline is False:
        ctx.chunk_analyze_chain.append(TargetVmaf())
    elif ctx.prototype_encoder.crf != -1:
        ctx.chunk_analyze_chain.append(PlainCrf())
    else:
        ctx.chunk_analyze_chain.append(PlainVbr())

    if len(ctx.chunk_analyze_chain) == 0:
        raise Exception("Failed to Create the analyze steps in analyzer_factory")

    return ctx


def setup_chunk_encoder(ctx, sequence):
    """
    Sets up the chunk encoder
    """

    # ugly imports here cuz "Circular import 🤓☝"
    from alabamaEncode.conent_analysis.chunk.final_encode_steps.plain import (
        PlainFinalEncode,
    )

    from alabamaEncode.conent_analysis.chunk.final_encode_steps.dynamic_target_vmaf_vbr import (
        DynamicTargetVmafVBR,
    )

    if ctx.dynamic_vmaf_target:
        ctx.chunk_encode_class = DynamicTargetVmaf()
    elif ctx.dynamic_vmaf_target_vbr:
        ctx.chunk_encode_class = DynamicTargetVmafVBR()
    elif ctx.multi_res_pipeline:
        print("Starting the multi res pipeline")
        ctx.chunk_encode_class = EncodeMultiResFinals()
    else:
        ctx.chunk_encode_class = PlainFinalEncode()

    return ctx


def get_refine_steps(ctx) -> List[RefineStep]:
    steps = []
    if ctx.multi_res_pipeline:
        steps.append(MutliResTrellis())
        steps.append(MutliResPackage())

    return steps


async def run_sequence_pipeline(ctx, sequence):
    from alabamaEncode.conent_analysis.sequence.autocrop import do_autocrop
    from alabamaEncode.conent_analysis.sequence.args_tune import (
        tune_args_for_fdlty_or_apl,
    )
    from alabamaEncode.conent_analysis.sequence.denoise_filtering import setup_denoise
    from alabamaEncode.conent_analysis.sequence.encoding_tiles import setup_tiles
    from alabamaEncode.conent_analysis.sequence.scrape_hdr_meta import (
        scrape_hdr_metadata,
    )
    from alabamaEncode.conent_analysis.sequence.sequence_autograin import (
        setup_autograin,
    )
    from alabamaEncode.conent_analysis.sequence.taget_ssimdb import setup_ssimdb_target
    from alabamaEncode.conent_analysis.sequence.x264_tune import get_ideal_x264_tune

    pipeline = [
        setup_chunk_analyze_chain,
        setup_chunk_encoder,
        scrape_hdr_metadata,
        tune_args_for_fdlty_or_apl,
        do_autocrop,
        setup_tiles,
        setup_denoise,
        setup_autograin,
        setup_ssimdb_target,
        get_ideal_x264_tune,
    ]

    for func in pipeline:
        # if async run awaits, if not run normally
        ctx = (
            await func(ctx, sequence)
            if func.__code__.co_flags & 0x80
            else func(ctx, sequence)
        )
    return ctx



from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.path import PathAlabama


def parse_video_filters(ctx):
    """
    Sets up the video filters
    """

    vec = ctx.prototype_encoder.video_filters.split(",")

    if ctx.crop_string != "":
        vec.append(f"crop={ctx.crop_string}")

    if ctx.scale_string != "":
        vec.append(f"scale={ctx.scale_string}:flags=lanczos")

    if ctx.prototype_encoder.hdr is False and Ffmpeg.is_hdr(
        PathAlabama(ctx.input_file)
    ):
        print("Input Video is HDR but requested a SDR encode, auto-tonemapping")
        vec.append(Ffmpeg.get_tonemap_vf())

    vec = [string for string in vec if string != ""]

    ctx.prototype_encoder.video_filters = ",".join(vec)
    return ctx



import copy
import json
import os

from scenedetect import detect, AdaptiveDetector

from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.scene.chunk import ChunkObject
from alabamaEncode.scene.sequence import ChunkSequence


def scene_detect(
    input_file: str,
    cache_file_path: str,
    max_scene_length: int = 10,
    start_offset=-1,
    end_offset=-1,
    override_bad_wrong_cache_path=False,
    static_length=False,
    scene_merge=False,
    static_length_size=30,
) -> ChunkSequence:
    """
    :param static_length_size:  size of static length scenes in seconds
    :param static_length: instead of detection, create static length scenes
    :param override_bad_wrong_cache_path:
    :param start_offset:
    :param end_offset:
    :param input_file: input file
    :param cache_file_path: path that the cache will be saved to
    :param max_scene_length: max scene length in seconds,
     cut the scene in the middle recursively until max_scene_length is reached
    :param scene_merge: merge scenes, so they are at minimum max_scene_length
    :return:
    """
    cache_path_is_valid = cache_file_path is not None and cache_file_path != ""

    seq = ChunkSequence([])
    if cache_path_is_valid and os.path.exists(cache_file_path):
        seq.load_json(open(cache_file_path).read())

        # Ensure input file matches the cached sequence
        if seq.input_file != input_file and not override_bad_wrong_cache_path:
            raise Exception(
                f"Video ({input_file}) != ({seq.input_file}) does not match scene cache ({cache_file_path}),"
                f" please correct this (wrong temp folder?), use --override_bad_wrong_cache_path to override."
            )
        return seq

    print("Running scene detection")

    scene_list = None

    untouched_scene_list = cache_file_path + ".untouched"
    if cache_path_is_valid:
        if os.path.exists(untouched_scene_list):
            print("Found untouched scene cache... loading")
            scene_list = json.load(open(untouched_scene_list))

    framerate: float = Ffmpeg.get_video_frame_rate(PathAlabama(input_file))
    width: int = Ffmpeg.get_width(PathAlabama(input_file))
    height: int = Ffmpeg.get_height(PathAlabama(input_file))

    static_length_size = int(static_length_size * framerate)

    if scene_list is None:
        if static_length:
            # create static length scenes
            total_size = Ffmpeg.get_frame_count_fast(PathAlabama(input_file))
            print(f"Using static length scenes of {static_length_size} frames")
            scene_list = [
                (i, min(i + static_length_size, total_size))
                for i in range(0, total_size, static_length_size)
            ]
        else:
            scene_list = detect(
                video_path=input_file,
                detector=AdaptiveDetector(
                    window_width=10, adaptive_threshold=2.5, min_content_val=12
                ),
                show_progress=True,
            )
            scene_list = [
                (scene[0].get_frames(), scene[1].get_frames()) for scene in scene_list
            ]

        if cache_path_is_valid:
            json.dump(scene_list, open(untouched_scene_list, "w"))

    max_length_frames = int(max_scene_length * framerate)
    if scene_merge and not static_length:
        new_scene_list = copy.deepcopy(scene_list)
        i = 1
        while i < len(new_scene_list):
            prev_start, prev_end = new_scene_list[i - 1]
            start, end = new_scene_list[i]
            if end - prev_start < 2 * max_length_frames:
                new_scene_list[i - 1] = (prev_start, end)
                print(
                    f"Merging scenes {i - 1} and {i} into {prev_start} - {end} ({end - prev_start} frames)"
                )
                del new_scene_list[i]
            else:
                i += 1

        scene_list = new_scene_list

    seq.input_file = input_file

    seq.chunks = []

    max_duration_frames = int(max_scene_length * framerate)

    # iterate through each scene detected in the video
    for scene in scene_list:
        start_frame, end_frame = scene
        duration = end_frame - start_frame

        # if the scene duration is shorter than max_scene_length, add it as a chunk to the sequence
        if duration <= max_duration_frames or static_length or scene_merge:
            seq.chunks.append(
                ChunkObject(
                    start_frame,
                    end_frame,
                    path=input_file,
                    framerate=framerate,
                    width=width,
                    height=height,
                )
            )
        else:
            # otherwise, split the scene into multiple chunks that are shorter than max_scene_length
            num_chunks = int(duration / max_duration_frames) + 1
            chunk_duration = duration / num_chunks
            for j in range(num_chunks):
                start = int(start_frame + j * chunk_duration)
                end = int(start_frame + (j + 1) * chunk_duration)
                if j == num_chunks - 1:
                    end = end_frame  # add any remaining frames to the last chunk
                seq.chunks.append(
                    ChunkObject(
                        start,
                        end,
                        path=input_file,
                        framerate=framerate,
                        width=width,
                        height=height,
                    )
                )

    if len(seq.chunks) == 0:
        print("Scene detection failed, falling back to a single chunk")
        seq.chunks.append(
            ChunkObject(
                0,
                Ffmpeg.get_frame_count(PathAlabama(input_file)),
                path=input_file,
                framerate=framerate,
                width=width,
                height=height,
            )
        )

    for i, chunk in enumerate(seq.chunks):
        chunk.chunk_index = i

    # cut any scenes that are in the offsets if a scene is at the border of the offset, cut it, so it fits
    if start_offset != -1 or end_offset != -1:
        new_chunks = []
        current_position = 0
        total_duration = Ffmpeg.get_video_length(PathAlabama(input_file))
        end_offset = total_duration - end_offset if end_offset != -1 else total_duration

        kept_chunks = []
        discarded_chunks = []

        for chunk in seq.chunks:
            # Calculate the chunk duration in seconds
            chunk_duration: float = chunk.get_lenght()

            # Calculate the start and end positions of the chunk
            chunk_start = current_position
            chunk_end = current_position + chunk_duration

            # Check if the chunk is entirely outside the bounds
            if (start_offset != -1 and chunk_end <= start_offset) or (
                end_offset != -1 and chunk_start >= end_offset
            ):
                # Chunk is entirely outside the bounds, discard it
                current_position += chunk_duration
                discarded_chunks.append(chunk.chunk_index)
                # print(f'Discarding chunk {chunk.chunk_index}')
                continue

            # Check if the chunk is entirely within the bounds
            if (start_offset == -1 or chunk_start >= start_offset) and (
                end_offset == -1 or chunk_end <= end_offset
            ):
                # Chunk is entirely within the bounds, include it in the new chunks
                kept_chunks.append(chunk.chunk_index)
                # print(f'Keeping chunk {chunk.chunk_index}')
                new_chunks.append(chunk)
            else:
                # Adjust the chunk start and end positions based on the offsets
                if start_offset != -1 and chunk_start < start_offset:
                    # Chunk starts before the start offset, adjust the start position
                    chunk_start = start_offset

                if end_offset != -1 and chunk_end > end_offset:
                    # Chunk ends after the end offset, adjust the end position
                    chunk_end = end_offset

                # Calculate the new chunk duration and frame indexes
                new_chunk_duration = chunk_end - chunk_start
                new_start_frame = int(
                    chunk.first_frame_index
                    + (chunk_start - current_position) * framerate
                )
                new_end_frame = int(new_start_frame + new_chunk_duration * framerate)

                print(
                    f"Cutting chunk {chunk.chunk_index} to {new_start_frame} - {new_end_frame}"
                )

                # Create a new chunk object with the updated values
                new_chunk = ChunkObject(
                    new_start_frame,
                    new_end_frame,
                    path=input_file,
                    framerate=framerate,
                    width=width,
                    height=height,
                )
                new_chunks.append(new_chunk)

            # Update the current position for the next chunk
            current_position = chunk_end

        if len(kept_chunks) == 0:
            raise Exception(
                "ERROR No chunks kept after cutting offsets, did you set the parameters correctly?"
            )

        print(f"keeping chunks {min(kept_chunks)}-{max(kept_chunks)}")
        print(f"discarding chunks {min(discarded_chunks)}-{max(discarded_chunks)}")

        # Replace the old list of chunks with the new one
        seq.chunks = new_chunks

    # test to see if any frames overlap
    # for i in range(len(seq.chunks) - 1):
    #     if seq.chunks[i].last_frame_index >= seq.chunks[i + 1].first_frame_index:
    #         print(
    #             f"Chunks {seq.chunks[i].chunk_index} and {seq.chunks[i + 1].chunk_index} overlap, "
    #             f"this should not happen"
    #         )
    # fix the overlap by cutting the first chunk
    # seq.chunks[i].last_frame_index = seq.chunks[i + 1].first_frame_index - 1

    # test to see if any frames are missing
    # for i in range(len(seq.chunks) - 1):
    #     if seq.chunks[i].last_frame_index + 1 != seq.chunks[i + 1].first_frame_index:
    #         print(
    #             f"Chunks {seq.chunks[i].chunk_index} and {seq.chunks[i + 1].chunk_index} are missing frames, "
    #             f"this should not happen"
    #         )
    # fix the missing frames by adding them to the first chunk
    # seq.chunks[i].last_frame_index = seq.chunks[i + 1].first_frame_index - 1

    # add chunk indexes, this has to be done after because scenes can be split, and keeping track of the index
    # would be painful
    for i, chunk in enumerate(seq.chunks):
        chunk.chunk_index = i

    # print all scenes
    # for i, chunk in enumerate(seq.chunks):
    #     print(
    #         f"Scene {i}: {chunk.first_frame_index} - {chunk.last_frame_index} ({chunk.get_lenght():.2f}s)"
    #     )

    if cache_path_is_valid:
        open(cache_file_path, "w").write(seq.dump_json())

    print(f"Detected {len(seq)} scenes")
    return seq



import json
import os

from matplotlib import pyplot as plt

from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.scene.chunk import ChunkObject

if __name__ == "__main__":
    chunk = ChunkObject(
        path="/mnt/data/objective-1-fast/Netflix_RollerCoaster_1280x720_60fps_8bit_420_60f.y4m"
    )

    features = Ffmpeg.get_content_features(chunk)

    print(json.dumps(features, indent=4))

    # on individual plots, plot every feature, assuming that features is a dict with the key as the frame index,
    # and the value is a dict of features
    # "0": {
    #     "entropy.entropy.normal.Y": "7.232096",
    #     "entropy.normalized_entropy.normal.Y": "0.723210",
    #     ...
    # },
    # "1": {
    #     "entropy.entropy.normal.Y": "7.226429",
    #     "entropy.normalized_entropy.normal.Y": "0.722643",
    #     "entropy.entropy.normal.U": "5.048188",
    #       ...
    # },
    # "2": {
    # etc

    env = "./ai_feature_extract"
    env = os.path.abspath(env)
    if not os.path.exists(env):
        os.mkdir(env)

    plot_dir = os.path.join(env, "plots")

    if not os.path.exists(plot_dir):
        os.mkdir(plot_dir)

    plots = []

    # Organize features by their names
    organized_features = dict()
    for frame_index, feature_dict in features.items():
        for feature_name, feature_value in feature_dict.items():
            if feature_name not in organized_features:
                organized_features[feature_name] = list()
            organized_features[feature_name].append(
                (int(frame_index), float(feature_value))
            )

    # Plot all features on their individual plots
    for feature_name, values in organized_features.items():
        x_values, y_values = zip(
            *values
        )  # Unpack frame indices and respective feature values

        plt.figure(figsize=(10, 5))
        plt.scatter(
            x_values, y_values
        )  # Consider using plt.scatter for individual points
        plt.title(f"Plot for {feature_name}")
        plt.grid()
        plt.xlabel("Frame Index")
        plt.ylabel(f"{feature_name}")
        # save plot
        plt.savefig(os.path.join(plot_dir, f"{feature_name}.png"))
        plots.append(os.path.join(plot_dir, f"{feature_name}.png"))
        plt.clf()
        plt.close()

    # create mosaic
    run_cli(
        f"montage {' '.join(plots)} -geometry +0+0  {os.path.join(plot_dir, 'mosaic.png')}"
    )



import os


class PathAlabama:
    def __init__(self, path: str):
        self.path = path

    def get(self):
        return self.path

    def get_safe(self):
        return f'"{self.path}"'

    def exists(self):
        return os.path.exists(self.path)

    def size_bytes(self):
        return os.path.getsize(self.path)

    def __str__(self):
        return self.get_safe()

    def check_video(self) -> bool:
        """
        Checks if the file exists and is a valid media file
        :return: True if the file exists and is a valid media file
        """
        if not self.exists():
            raise FileNotFoundError(f"File {self.path} does not exist")
        if self.size_bytes() < 10:
            raise ValueError(f"File {self.path} is too small to be a valid media file")
        return True



class MetricException(Exception):
    """Base class for exceptions in this module."""

    pass


class VmafException(MetricException):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class Ssimu2Exception(MetricException):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



from typing import List

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution


class EncoderX265(Encoder):

    def get_pretty_name(self) -> str:
        return "X265"

    def get_codec(self) -> Codec:
        return Codec.h265

    def get_version(self) -> str:
        return (
            run_cli(f"{get_binary('x265')} --help")
            .get_output()
            .split("\n")[0]
            .replace("x265 [info]: HEVC encoder version ", "")
            .strip()
        )

    def get_encode_commands(self) -> List[str]:
        if self.speed == 0:
            self.speed = 1

        kommand = (
            f" {self.get_ffmpeg_pipe_command()} | {get_binary('x265')} --input - --y4m "
        )

        if self.hdr:
            kommand += f" --output-depth {self.bit_override} "
            colormatrix = self.matrix_coefficients
            if colormatrix == "bt2020-ncl":
                colormatrix = "bt2020nc"
            chromeloc = self.chroma_sample_position
            if chromeloc == "topleft":
                chromeloc = "0"
            kommand += (
                f" --colorprim={self.color_primaries} --transfer={self.transfer_characteristics} "
                f"--colormatrix={colormatrix} --chromaloc={chromeloc} "
                f"--max-cll {self.maximum_content_light_level},{self.maximum_frame_average_light_level}"
            )
            kommand += f" --hdr10 "
            # if self.svt_master_display:
            #     kommand += f' --mastering-display "{self.svt_master_display}" '

        kommand += " --no-scenecut "

        kommand += f" --keyint={self.keyint} "

        match self.rate_distribution:
            case EncoderRateDistribution.CQ:
                kommand += f" --crf={self.crf}"
            case _:
                raise Exception(
                    f"FATAL: rate distribution {self.rate_distribution} not supported"
                )

        match self.speed:
            case 9:
                kommand += " --preset=ultrafast"
            case 8:
                kommand += " --preset=superfast"
            case 7:
                kommand += " --preset=veryfast"
            case 6:
                kommand += " --preset=faster"
            case 5:
                kommand += " --preset=fast"
            case 4:
                kommand += " --preset=medium"
            case 3:
                kommand += " --preset=slow"
            case 2:
                kommand += " --preset=slower"
            case 1:
                kommand += " --preset=veryslow"
            case 0:
                kommand += " --preset=placebo"

        if self.passes == 2:
            raise Exception("FATAL: 2 pass encoding not supported")
        elif self.passes == 1:
            hevc_file = self.output_path.replace(".mkv", ".hevc")
            remux_command = (
                f'{get_binary("mkvmerge")} -o "{self.output_path}" "{hevc_file}"'
            )
            del_commnad = f'rm "{hevc_file}"'
            return [f'{kommand} -o "{hevc_file}"', remux_command, del_commnad]

    def get_chunk_file_extension(self) -> str:
        return ".mkv"

    def supports_float_crfs(self) -> bool:
        return True


if __name__ == '__main__':
    x265 = EncoderX265()
    print(x265.get_pretty_name())
    print(x265.get_codec())
    print(x265.get_version())
    print(x265.get_chunk_file_extension())
    print(x265.supports_float_crfs())


import subprocess

from alabamaEncode.core.util.abort_controler import AbortControler
from alabamaEncode.core.util.yuv import Reader
from alabamaEncode.scene.chunk import ChunkObject


def get_yuv_frame_stream(chunk: ChunkObject, frame_callback, vf: str= "", abort_controler: AbortControler = None):
    command = chunk.create_chunk_ffmpeg_pipe_command(video_filters=vf, bit_depth=8)

    ffmpeg_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

    yuv_reader = Reader(frame_callback)

    while ffmpeg_process.poll() is None:
        if abort_controler:
            if abort_controler.aborted:
                break
        data = ffmpeg_process.stdout.read(1024 * 1024)
        if data:
            yuv_reader.decode(data)



import os
import re

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.scene.chunk import ChunkObject


def get_video_ssim(
    distorted_path,
    in_chunk: ChunkObject = None,
    print_output=False,
    get_db=False,
    video_filters="",
):
    if not os.path.exists(in_chunk.path) or not os.path.exists(distorted_path):
        raise FileNotFoundError(
            f"File {in_chunk.path} or {distorted_path} does not exist"
        )
    null_ = in_chunk.create_chunk_ffmpeg_pipe_command(video_filters=video_filters)

    null_ += f" | {get_binary('ffmpeg')} -hide_banner -i - -i {distorted_path} -filter_complex ssim -f null -"

    result_string = run_cli(null_).get_output()
    if print_output:
        print(result_string)
    try:
        # Regex to get avrage score out of:
        # [Parsed_ssim_0 @ 0x55f6705fa200] SSIM Y:0.999999 (0.000000 dB) U:0.999999 (0.000000 dB)
        # V:0.999999 (0.000000 dB) All:0.999999 (0.000000)
        match = re.search(r"All:([\d.]+) \(([\d.]+)", result_string)
        ssim_score = float(match.group(1))
        ssim_db = float(match.group(2))

        if get_db is True:
            return ssim_score, ssim_db
        else:
            return ssim_score
    except AttributeError:
        print(f"Failed getting ssim comparing {distorted_path} agains {in_chunk.path}")
        print(null_)
        return 0



import time


class Timer:
    def __init__(self):
        self.timers = {}

    def start(self, name: str):
        self.timers[name] = time.time()

    def stop(self, name: str):
        if name not in self.timers:
            raise Exception("Timer not started")
        self.timers[name] = time.time() - self.timers[name]
        return self.timers[name]

    def finish(self, loud=False):
        if loud:
            print("timers:")
            for key in self.timers:
                print(f"{key}: {self.timers[key]}s")
        return self.timers



import os

from alabamaEncode.conent_analysis.chunk.chunk_analyse_step import (
    ChunkAnalyzePipelineItem,
)
from alabamaEncode.conent_analysis.opinionated_vmaf import (
    convexhull_get_resolutions,
    convexhull_get_crf_range,
)
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.calculate import get_metric_from_stats
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.chunk import ChunkObject


class EncodeMultiResCandidates(ChunkAnalyzePipelineItem):
    def run(self, ctx, chunk: ChunkObject, enc: Encoder) -> Encoder:
        crf_low, crf_high = convexhull_get_crf_range(enc.get_codec())
        crf_range = list(range(crf_low, crf_high, 2))
        resolutions = convexhull_get_resolutions(enc.get_codec())
        crf_range.reverse()  # encode the probes fast to slow, not necessary but why not
        resolutions.reverse()

        ctx.vmaf_reference_display = "FHD"
        vmaf_options = ctx.get_vmaf_options()

        probe_folder_path = (
            os.path.join(os.path.dirname(chunk.chunk_path), f"{chunk.chunk_index}")
            + os.path.sep
        )
        os.makedirs(probe_folder_path, exist_ok=True)

        ogspeed = enc.speed
        enc.speed = 13

        def log(_str):  # macro
            ctx.log(
                f"{chunk.log_prefix()}{_str}",
                category="multi_res",
            )

        for res in resolutions:
            vf = ctx.prototype_encoder.video_filters.split(",")
            scale_str = f"scale={res}:flags=lanczos"

            # replace the first scale filter if it exists
            for i in range(len(vf)):
                if ":flags=lanczos" in vf[i]:
                    vf[i] = scale_str
                    break

            # if there is no scale filter, add it to the end
            if scale_str not in ",".join(vf):
                vf.append(scale_str)

            vf = [f for f in vf if f != ""]

            enc.video_filters = ",".join(vf)

            for crf in crf_range:
                enc.crf = crf

                res_name_short = res.split(":")[0]

                enc.output_path = os.path.join(
                    probe_folder_path,
                    f"{chunk.chunk_index}.{res_name_short}.{crf}{enc.get_chunk_file_extension()}",
                )

                if ctx.get_kv().exists(
                    "multi_res_candidates",
                    f"{chunk.chunk_index}_{res_name_short}_{crf}",
                ):
                    continue

                stats: EncodeStats = enc.run(
                    metric_params=vmaf_options, metric_to_calculate=Metric.VMAF
                )

                vmaf = get_metric_from_stats(
                    stats, statistical_representation=ctx.vmaf_target_representation
                )

                log(f"Res: {res} VMAF: {vmaf} CRF: {crf} Bitrate: {stats.bitrate}")

                ctx.get_kv().set(
                    "multi_res_candidates",
                    f"{chunk.chunk_index}_{res_name_short}_{crf}",
                    {
                        "vmaf": vmaf,
                        "crf": crf,
                        "bitrate": stats.bitrate,
                        "file": enc.output_path,
                        "res": res,
                    },
                )

        enc.speed = ogspeed
        return enc



def parse_resolution_presets(ctx):
    """
    Computes the resolution presets
    """
    if ctx.resolution_preset != "" and ctx.scale_string == "":
        # 4k 1440p 1080p 768p 720p 540p 480p 360p
        match ctx.resolution_preset:
            case "4k":
                ctx.scale_string = "3840:-2"
            case "1440p":
                ctx.scale_string = "2560:-2"
            case "1080p":
                ctx.scale_string = "1920:-2"
            case "768p":
                ctx.scale_string = "1366:-2"
            case "720p":
                ctx.scale_string = "1280:-2"
            case "540p":
                ctx.scale_string = "960:-2"
            case "480p":
                ctx.scale_string = "854:-2"
            case "432p":
                ctx.scale_string = "768:-2"
            case "360p":
                ctx.scale_string = "640:-2"
            case "240p":
                ctx.scale_string = "480:-2"
            case _:
                raise ValueError(
                    f'Cannot interpret resolution preset "{ctx.resolution_preset}", refer to the help command'
                )
    return ctx



from setuptools import setup, find_packages

setup(
    name="alabamaEncoder",
    version="0.5.4",
    packages=find_packages(),
    install_requires=[
        "scenedetect",
        "tqdm",
        "celery",
        "redis",
        "psutil",
        "opencv-contrib-python",
        "requests",
        "torf",
        "websockets",
        "scipy",
        "numpy",
        "scikit-image",
        "argparse_range",
        "matplotlib",
    ],
    entry_points="""
      [console_scripts]
      alabamaEncoder=alabamaEncode.cli.__main__:main
      """,
    python_requires=">=3.12.3",
)



from typing import List

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution


class EncoderVaapiH264(Encoder):
    def get_pretty_name(self) -> str:
        return "VAAPI_H264"

    def get_codec(self) -> Codec:
        return Codec.h264

    def get_encode_commands(self) -> List[str]:
        self.bit_override = 8  # TODO: check vaapi profile support
        vec = []
        vec.append(self.get_ffmpeg_pipe_command())
        vec.append("|")
        vec.append(get_binary("ffmpeg"))
        vec.append(" -hide_banner -y -i -")
        vec.append(f"-c:v h264_vaapi")
        vec.append(f"-vaapi_device /dev/dri/renderD128")

        vec.append(f"-g {self.keyint}")

        match self.rate_distribution:
            case EncoderRateDistribution.CQ:
                vec.append(f"-qp {self.crf}")
                vec.append("-rc_mode 1")
            case EncoderRateDistribution.VBR:
                vec.append(f"-b:v {self.bitrate}k")
                vec.append("-rc_mode 3")
            case _:
                raise NotImplementedError

        vec.append(f"-quality 0")
        vec.append(f"-compression_level 0")
        vec.append(f"-profile high")

        # hdr
        if self.hdr:
            vec.append(f"-color_primaries {self.color_primaries}")
            vec.append(f"-color_trc {self.transfer_characteristics}")
            vec.append(f"-colorspace {self.matrix_coefficients}")
            vec.append(f"-master_display {self.svt_master_display}")
            vec.append(
                f"-max_cll {self.maximum_content_light_level},{self.maximum_frame_average_light_level}"
            )

        if self.bit_override == 10:
            vec.append('-vf "format=p010,hwupload"')
        elif self.bit_override == 8:
            vec.append('-vf "format=nv12,hwupload"')

        # output
        vec.append(f'"{self.output_path}"')

        return [" ".join(vec)]

    def get_chunk_file_extension(self) -> str:
        return ".avi"

    def get_version(self) -> str:
        return "N/A"




import os
import pickle
from concurrent.futures import ThreadPoolExecutor
from typing import List

from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.scene.chunk import ChunkObject
from alabamaEncode.scene.sequence import ChunkSequence


def calulcate_ssimdb(bitrate: int, chunk: ChunkObject, dbs: List[float], ctx):
    """
    Calculates the ssim dB for a chunk and appends it to the dbs list
    :param ctx: AlabamaContext
    :param bitrate: bitrate in kbps
    :param chunk: chunk to calculate ssim dB for
    :param dbs: The list to append the ssim dB to
    """
    enc = ctx.get_encoder()
    enc.chunk = chunk
    enc.speed = 6
    enc.passes = 3
    enc.grain_synth = 0
    enc.rate_distribution = EncoderRateDistribution.VBR
    enc.threads = 1
    enc.bitrate = bitrate
    enc.output_path = (
        f"{ctx.temp_folder}adapt/bitrate/ssim_translate/{chunk.chunk_index}"
        f"{enc.get_chunk_file_extension()}"
    )
    try:
        stats: EncodeStats = enc.run(calcualte_ssim=True)
    except Exception as e:
        print(f"Failed to calculate ssim dB for {chunk.chunk_index}: {e}")
        return
    ctx.log(f"[{chunk.chunk_index}] {bitrate} kbps -> {stats.ssim_db} ssimdb")
    dbs.append(stats.ssim_db)


def get_target_ssimdb(bitrate: int, ctx, chunk_sequence) -> float:
    """
    Since in the AutoBitrate we are targeting ssim dB values, we need to somehow translate vmaf to ssim dB
    :param ctx: AlabamaContext
    :param chunk_sequence: ChunkSequence
    :param bitrate: bitrate in kbps
    :return: target ssimdb
    """
    print(f"Getting target ssim dB for {bitrate} kbps")
    cache_path = f"{ctx.temp_folder}/adapt/bitrate/ssim_translate/{bitrate}.pl"
    if os.path.exists(cache_path):
        target_ssimdb = pickle.load(open(cache_path, "rb"))
        print(f"cached ssim dB for {bitrate}: {target_ssimdb}dB")
        return target_ssimdb
    dbs = []
    os.makedirs(f"{ctx.temp_folder}/adapt/bitrate/ssim_translate", exist_ok=True)

    chunks = chunk_sequence.get_test_chunks_out_of_a_sequence(7)

    with ThreadPoolExecutor(max_workers=3) as executor:
        for chunk in chunks:
            executor.submit(calulcate_ssimdb, bitrate, chunk, dbs, ctx)
        executor.shutdown()

    target_ssimdb = sum(dbs) / len(dbs)

    print(f"Avg ssim dB for {bitrate}Kbps: {target_ssimdb}dB")
    pickle.dump(target_ssimdb, open(cache_path, "wb"))
    return target_ssimdb


def setup_ssimdb_target(ctx: AlabamaContext, sequence: ChunkSequence):
    if all(
        [
            ctx.crf_based_vmaf_targeting is False,
            ctx.vbr_perchunk_optimisation,
        ]
    ):
        if ctx.prototype_encoder.bitrate is None:
            print("No bitrate set, cannot set SSIMDB target, set --bitrate")
            quit()
        print("Using SSIMDB target")
        ctx.ssim_db_target = get_target_ssimdb(
            ctx.prototype_encoder.bitrate, ctx=ctx, chunk_sequence=sequence
        )

    return ctx



#!/usr/bin/python
import asyncio
import atexit
import os
import sys
import time

from alabamaEncode.cli.cli_setup.autopaths import auto_output_paths
from alabamaEncode.cli.cli_setup.cli_args import read_args
from alabamaEncode.cli.cli_setup.paths import parse_paths
from alabamaEncode.cli.cli_setup.ratecontrol import parse_rd
from alabamaEncode.cli.cli_setup.res_preset import parse_resolution_presets
from alabamaEncode.cli.cli_setup.validate_files import validate_input
from alabamaEncode.cli.cli_setup.video_filters import parse_video_filters
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.extras.auto_thumbnailer import AutoThumbnailer
from alabamaEncode.core.job import AlabamaEncodingJob
from alabamaEncode.parallel_execution.celery_app import app
from alabamaEncode.parallel_execution.worker import worker

if sys.version_info < (3, 12, 3):
    print("Python 3.12.3 or a more recent version is required.")
    sys.exit(1)

runtime = -1
runtime_file = ""
lock_file_path = ""


def run_pipeline(ctx):
    creation_pipeline = [
        auto_output_paths,
        parse_paths,
        validate_input,
        parse_rd,
        parse_resolution_presets,
        parse_video_filters,
    ]
    for pipeline_item in creation_pipeline:
        ctx = pipeline_item(ctx)
    return ctx


def fmt_time(seconds) -> str:
    # Convert seconds to hours and remaining seconds
    hours, remainder = divmod(seconds, 3600)

    # If at least one hour has passed
    if hours:
        return f"{int(hours)} hour(s) {int(remainder)} second(s)"
    else:
        return f"{int(seconds)} second(s)"


@atexit.register
def at_exit():
    global runtime
    global runtime_file
    if runtime != -1:
        current_session_runtime = time.time() - runtime

        saved_runtime = 0
        if os.path.exists(runtime_file):
            with open(runtime_file) as f:
                saved_runtime = float(f.read())
        print(
            f"Current Session Runtime: {fmt_time(current_session_runtime)},"
            f" Runtime From Previous Sessions: {fmt_time(saved_runtime)},"
            f" Total Runtime: {fmt_time(current_session_runtime + saved_runtime)}"
        )

        try:
            with open(runtime_file, "w") as f:
                f.write(str(current_session_runtime + saved_runtime))
        except FileNotFoundError:
            pass
        if os.path.exists(lock_file_path):
            os.remove(lock_file_path)


def main():
    """
    Main entry point
    """
    if os.name == "nt":
        print("Windows is not supported")
        quit()

    global runtime
    runtime = time.time()

    ctx: AlabamaContext = AlabamaContext()

    ctx, args = read_args(ctx)

    match args.command:
        case "clear":
            print("Clearing celery queue")
            app.control.purge()
            quit()
        case "worker":
            worker(args.workers)
        case "autothumbnailer":
            if not os.path.exists(args.input):
                print(f"File {args.input} does not exist")
                quit()

            thumbnailer = AutoThumbnailer()
            thumbnailer.calc_face = args.detect_faces
            thumbnailer.generate_previews(
                input_file=args.input, output_folder=os.path.dirname(args.input)
            )
            quit()

    ctx = run_pipeline(ctx)

    global runtime_file
    global lock_file_path
    runtime_file = os.path.join(ctx.temp_folder, "runtime.txt")
    lock_file_path = os.path.join(ctx.output_folder, "alabama.lock")

    if os.path.exists(lock_file_path):
        print(
            "Lock file exists, are you sure another instance is not encoding in this folder? "
            "if not delete the lock file and try again"
        )
        quit()
    else:
        open(lock_file_path, "w").close()

    output_lock = os.path.join(ctx.temp_folder, "output.lock")
    if not os.path.exists(output_lock):
        with open(output_lock, "w") as f:
            f.write(ctx.raw_input_file)
    else:
        output_file_from_lock = open(output_lock).read()
        if output_file_from_lock != ctx.raw_input_file:
            print(
                f"Output file from lock file {output_file_from_lock} does not match output file "
                f"from ctx {ctx.raw_input_file}"
            )
            quit()

    if ctx.offload_server != "":
        print("Offloading to remote server")
        auth_token = os.environ.get("AUTH_BEARER_TOKEN", "")
        if auth_token == "":
            print("No AUTH_BEARER_TOKEN set, exiting")
            sys.exit(1)
        import requests
        import json

        headers = {"Authorization": f"Bearer {auth_token}"}
        data = json.dumps(ctx.to_json())
        requests.post(f"{ctx.offload_server}/jobs", data=data, headers=headers)

    job = AlabamaEncodingJob(ctx)

    asyncio.run(job.run_pipeline())  # this runs the whole encoding process

    quit()


if __name__ == "__main__":
    main()



import hashlib
import json
import multiprocessing
import os

import cv2
import numpy as np
import scipy
from skimage.metrics import structural_similarity as ssim
from tqdm import tqdm

from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.bin_utils import (
    get_binary,
    check_ffmpeg_libraries,
    check_bin,
)
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.util.get_yuv_stream import get_yuv_frame_stream
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.scene.chunk import ChunkObject


def calculate_saliency_score(image):
    # Compute the saliency map
    (success, saliencyMap) = (
        cv2.saliency.StaticSaliencySpectralResidual_create().computeSaliency(image)
    )

    # Check for computation errors
    if not success:
        return 0

    # Convert the saliency map to a NumPy array
    saliencyMap = saliencyMap.astype("float32")

    # Calculate the average saliency score
    average_score = cv2.mean(saliencyMap)[0]

    return average_score


def calculate_rule_of_thirds(image, gray):
    height, width, _ = image.shape
    thirds_x = [width // 3, 2 * width // 3]
    thirds_y = [height // 3, 2 * height // 3]

    _, thresholded = cv2.threshold(gray, 128, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)

    key_points = []
    contours, _ = cv2.findContours(
        thresholded, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    for contour in contours:
        M = cv2.moments(contour)
        if M["m00"] != 0:
            cX = int(M["m10"] / M["m00"])
            cY = int(M["m01"] / M["m00"])
            key_points.append((cX, cY))

    third_score = 0
    for cX, cY in key_points:
        if any(abs(cX - tx) <= width // 20 for tx in thirds_x) and any(
            abs(cY - ty) <= height // 20 for ty in thirds_y
        ):
            third_score += 1

    return third_score / len(key_points) if key_points else 0


def compute_shot_symmetry_score(gray):

    # Split the image into left and right halves
    height, width = gray.shape
    left_half = gray[:, : width // 2]
    right_half = gray[:, width // 2 :]

    # Flip the right half horizontally for comparison
    right_half_flipped = cv2.flip(right_half, 1)

    # Compute the SSIM between the left half and flipped right half
    symmetry_score, _ = ssim(left_half, right_half_flipped, full=True)

    return symmetry_score


def compute_depth_of_field_score(gray):

    # Divide the image into regions (e.g., top, middle, bottom)
    height, width = gray.shape
    regions = [
        gray[: height // 3, :],
        gray[height // 3 : 2 * height // 3, :],
        gray[2 * height // 3 :, :],
    ]

    # Calculate the variance of the Laplacian for each region
    laplacian_vars = [cv2.Laplacian(region, cv2.CV_64F).var() for region in regions]

    # Calculate the standard deviation of the variances
    std_dev_laplacian = np.std(laplacian_vars)

    # Normalize the score (higher score indicates more variation in focus, hence shallower depth of field)
    dof_score = std_dev_laplacian / max(laplacian_vars)

    return dof_score


def face_score(frame) -> float:
    # detect faces, their expressions, and give it a float score
    return 0


def process_frame_worker(yuv_frame_buffer, h, w, count, calc_face=False):
    frame = np.frombuffer(yuv_frame_buffer, dtype=np.uint8).reshape((h * 3 // 2, w))
    frame = cv2.cvtColor(frame, cv2.COLOR_YUV2BGR_I420)
    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    aa = {
        "contrast": gray.std(),
        #   "rule_of_thirds": calculate_rule_of_thirds(frame, gray),
        "saliency": calculate_saliency_score(frame),
        "blurriness": cv2.Laplacian(frame, cv2.CV_64F).var(),
        #   "saturation": cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)[:, :, 1].mean(),
        "dof": compute_depth_of_field_score(gray),
        #   "symmetry": compute_shot_symmetry_score(gray),
        "index": count,
    }
    if calc_face:
        aa["face_score"] = face_score(frame)
    return aa


class AutoThumbnailer:
    def __init__(self):
        self.pbar = None

        self.frame_data = []
        self.pool = multiprocessing.Pool(6)
        self.calc_face = False

    def generate_previews(self, input_file: str, output_folder: str):
        chunk = ChunkObject(path=input_file)

        random_part = hashlib.sha1(input_file.encode("utf-8")).hexdigest()[:5]

        frame_data_path = f"{output_folder}/thumbnail_frame_data_{random_part}.json"

        try:
            if os.path.exists(frame_data_path):
                with open(frame_data_path, "r") as f:
                    self.frame_data = json.load(f)
                    print(f"Loaded {len(self.frame_data)} frames from cache")
        except:
            pass

        if len(self.frame_data) <= 0:
            total_length = Ffmpeg.get_frame_count(PathAlabama(chunk.path))
            self.pbar = tqdm(
                total=total_length, desc="Gathering thumbnail data", unit="frame"
            )

            get_yuv_frame_stream(
                chunk,
                frame_callback=self.process_frame,
                vf="\"scale=-2:min'(720,ih)':force_original_aspect_ratio=decrease\"",
            )

            # await all the tasks in the pool to be finished
            self.pool.close()
            self.pool.join()

            with open(frame_data_path, "w") as f:
                json.dump(self.frame_data, f)

            self.pbar.close()

            print(f"Processed {len(self.frame_data)} frames")

        if len(self.frame_data) == 0:
            raise Exception("No frames were processed")

        features_to_calc = [
            "blurriness",
            "saliency",
            # 'rule_of_thirds',
            "contrast",
            # 'saturation',
            "dof",
            # 'symmetry'
        ]

        if self.calc_face:
            features_to_calc.append("face_score")

        best_frames = self.get_top_frames(
            self.frame_data,
            num_peaks=9,
            feature_names=features_to_calc,
        )

        has_placebo = check_ffmpeg_libraries("libplacebo")
        has_jpegli = check_bin("cjpeg")
        for i, best_frame in tqdm(
            enumerate(best_frames), desc="Saving best frames", total=len(best_frames)
        ):
            chunk = ChunkObject(
                first_frame_index=best_frame,
                last_frame_index=best_frame + 1,
                path=input_file,
            )
            if has_placebo and has_jpegli:
                command = f"{get_binary('ffmpeg')} -init_hw_device vulkan -y {chunk.get_ss_ffmpeg_command_pair()} -frames:v 1 -vf 'hwupload,libplacebo=minimum_peak=2:percentile=99.6:tonemapping=spline:colorspace=bt709:color_primaries=bt709:gamut_mode=perceptual:color_trc=bt709:range=tv:gamma=1:format=yuv420p,hwdownload,format=yuv420p' -c:v png -f image2pipe - | {get_binary('cjpeg')} -q 95 -tune-psnr -optimize -progressive > \"{output_folder}/{i}.jpg\""
            elif has_placebo:
                command = f"{get_binary('ffmpeg')} -init_hw_device vulkan -y {chunk.get_ss_ffmpeg_command_pair()} -frames:v 1 -vf 'hwupload,libplacebo=minimum_peak=2:percentile=99.6:tonemapping=spline:colorspace=bt709:color_primaries=bt709:gamut_mode=perceptual:color_trc=bt709:range=tv:gamma=1:format=yuv420p,hwdownload,format=yuv420p' \"{output_folder}/{i}.png\""
            else:
                command = f"{get_binary('ffmpeg')} -y {chunk.get_ss_ffmpeg_command_pair()} -frames:v 1 \"{output_folder}/{i}.png\""
            run_cli(command).verify()

    def process_frame(self, yuv_frame):
        self.pool.apply_async(
            process_frame_worker,
            args=(
                yuv_frame.buffer,
                yuv_frame.headers["H"],
                yuv_frame.headers["W"],
                yuv_frame.count,
                self.calc_face,
            ),
            callback=self.collect_result,
        )

    def collect_result(self, result):
        self.frame_data.append(result)
        self.pbar.update()

    # def process_frame(self, yuv_frame):
    #     frame = np.frombuffer(yuv_frame.buffer, dtype=np.uint8).reshape((yuv_frame.headers['H'] * 3 // 2, yuv_frame.headers['W']))
    #     frame = cv2.cvtColor(frame, cv2.COLOR_YUV2BGR_I420)
    #     gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    #
    #     self.frame_data.append({
    #         "blurriness": cv2.Laplacian(frame, cv2.CV_64F).var(),
    #         "contrast": gray.std(),
    #         "saliency": calculate_saliency_score(frame),
    #         "index": yuv_frame.count,
    #         "rule_of_thirds": calculate_rule_of_thirds(frame, gray)
    #     })
    #     self.pbar.update()

    def normalize_and_combine(self, frame_data, feature_names):
        features = {
            feature: np.array([frame[feature] for frame in frame_data])
            for feature in feature_names
        }

        # Normalize each feature to [0, 1]
        normalized_features = {}
        for feature, values in features.items():
            min_val = values.min()
            max_val = values.max()
            if max_val - min_val != 0:  # Avoid division by zero
                normalized_features[feature] = (values - min_val) / (max_val - min_val)
            else:
                normalized_features[feature] = values - min_val

        # if any of the values contain nan's kick them out
        # normalized_features = {k: v for k, v in normalized_features.items() if not np.isnan(v).any()}

        # change any nans in the data into 0's
        for feature, values in normalized_features.items():
            normalized_features[feature] = np.nan_to_num(values)

        # Combine the normalized scores into a single score
        combined_score = np.zeros_like(list(normalized_features.values())[0])
        for feature, normalized_values in normalized_features.items():
            if feature == "contrast":
                normalized_values *= 0.7
            elif feature == "saliency":
                normalized_values *= 1.5
            elif feature == "blurriness":
                normalized_values *= 1.8
            elif feature == "saturation":
                normalized_values *= 0.4
            elif feature == "dof":
                normalized_values *= 0.9

            combined_score += normalized_values
        return combined_score

    # Step 2: Smooth the data to highlight peaks
    def smooth_data(self, combined_score):
        # print pre and post smoothing on top of eathcother in color
        # smoothed_score = scipy.signal.savgol_filter(combined_score, window_length=20, polyorder=2)
        # smoothed_score = scipy.ndimage.gaussian_filter1d(combined_score, sigma=100)
        # alt mode where you increate the smoothing factor based on lenght up to a point
        smoothing_factor = min(100, len(combined_score) // 10)
        smoothed_score = scipy.ndimage.gaussian_filter1d(
            combined_score, sigma=smoothing_factor
        )

        # import matplotlib.pyplot as plt
        #
        # plt.plot(combined_score, label="Combined Score")
        # plt.plot(smoothed_score, label="Smoothed Score")
        # plt.legend()
        # plt.show()

        return smoothed_score

    # Step 3: Find peaks in the smoothed data
    def find_peaks(self, smoothed_score):
        peaks, _ = scipy.signal.find_peaks(
            smoothed_score, distance=len(smoothed_score) // 20
        )  # Adjust distance for peak spacing

        # print the data with peeks overlayed
        # import matplotlib.pyplot as plt
        #
        # plt.plot(smoothed_score, label="Smoothed Score")
        # plt.plot(peaks, smoothed_score[peaks], "x", label="Peaks")
        # plt.legend()
        # plt.show()

        return peaks

    # Step 4: Select evenly spaced peaks or fallback to highest scores
    def select_evenly_spaced_peaks(self, peaks, num_peaks, combined_score):

        if not len(peaks) >= num_peaks:
            print(f"Detection only found {len(peaks)} good candidates, using them")
            return peaks
        else:
            step = len(peaks) // num_peaks
            evenly_spaced_peaks = peaks[::step][:num_peaks]
        # if len(peaks) >= num_peaks:
        #     step = len(peaks) // num_peaks
        #     evenly_spaced_peaks = peaks[::step][:num_peaks]
        # else:
        #     # Fallback: select frames with the highest combined scores
        #     top_indices = np.argsort(-combined_score)[:num_peaks]
        #     evenly_spaced_peaks = sorted(top_indices)
        return evenly_spaced_peaks

    # Main function to get top frame indices
    def get_top_frames(self, frame_data, feature_names, num_peaks=5):
        print("Processing frame data...")
        combined_score = self.normalize_and_combine(frame_data, feature_names)
        smoothed_score = self.smooth_data(combined_score)
        peaks = self.find_peaks(smoothed_score)
        selected_peaks = self.select_evenly_spaced_peaks(
            peaks, num_peaks, combined_score
        )
        selected_indices = [frame_data[peak]["index"] for peak in selected_peaks]

        print(f"Selected frames: {selected_indices}")
        return selected_indices


if __name__ == "__main__":
    AutoThumbnailer().generate_previews(
        "/home/kokoniara/Downloads/I Bought 6 PC “Speed Up” Tools to See if They Work [-G-DByczbWA].webm",
        ".",
    )
    # AutoThumbnailer().generate_previews(
    #     "/home/kokoniara/showsEncode/Silo (2023)/s1/e9/Silo.2023.S01E09.OPUS.AV1.1080p.webm",
    #     ".",
    # )



import re
from typing import List

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution


class EncoderVPX(Encoder):
    def get_codec(self) -> Codec:
        return Codec.vp9

    def get_pretty_name(self) -> str:
        if self.codec == "vp9":
            return "VPX_VP9"
        elif self.codec == "vp8":
            return "VPX_VP8"


    def __init__(self, codec):
        super().__init__()
        if codec != "vp9" and codec != "vp8":
            raise ValueError(f"Invalid encoder: {codec}")
        self.codec = codec

    def get_encode_commands(self) -> List[str]:
        cli = ""
        cli += self.get_ffmpeg_pipe_command()
        cli += " | "
        vec = []
        vec.append(f"{get_binary('vpxenc')} -")
        vec.append(f"--output={self.output_path}")
        vec.append(f"--codec={self.codec}")
        vec.append(f"--ivf")
        vec.append(f"--threads={self.threads}")
        vec.append(f"--good")
        vec.append(f"--kf-max-dist={self.keyint}")
        vec.append(f"--sharpness=2")

        if self.codec == "vp9":
            self.speed = min(self.speed, 9)
            vec.append(f"--bit-depth={self.bit_override}")
            if self.bit_override == 10:
                vec.append(f"--input-bit-depth=10")
                vec.append(f"--profile=2")
            vec.append(f"--aq-mode=2")
            vec.append(f"--lag-in-frames=25")
            vec.append(f"--auto-alt-ref=1")
            vec.append(f"--row-mt=1")
            vec.append(f"--enable-tpl=1")

        elif self.codec == "vp8":
            self.speed = min(self.speed, 16)
            if self.bit_override == 10:
                print("VP8 does not support 10 bit, setting to 8 bit")
                self.bit_override = 8
            if self.hdr:
                raise ValueError("VP8 does not support HDR")

        vec.append(f"--cpu-used={self.speed}")

        match self.rate_distribution:
            case EncoderRateDistribution.VBR:
                vec.append(f"--end-usage=vbr")
                vec.append(f"--target-bitrate={self.bitrate}")
            case EncoderRateDistribution.CQ:
                vec.append(f"--end-usage=q")
                vec.append(f"--cq-level={self.crf}")
            case EncoderRateDistribution.CQ_VBV:
                vec.append(f"--end-usage=cq")
                vec.append(f"--cq-level={self.crf}")
                vec.append(f"--target-bitrate={self.bitrate}")

        if self.passes == 1:
            vec.append(f"--passes=1")
            return [cli + " ".join(vec)]
        elif self.passes == 2:
            vec.append(f"--passes=2")
            vec.append(f"--fpf={self.output_path}.log")
            commands = []
            for i in range(1, self.passes + 1):
                commands.append(
                    cli + " ".join(vec) + f" --pass={i} --fpf={self.output_path}.log"
                )
            return commands

    def get_chunk_file_extension(self) -> str:
        return ".ivf"

    def get_version(self) -> str:
        # vp9    - WebM Project VP9 Encoder v1.12.0 (default)
        match = re.search(
            r"vp9\s+-\s+WebM Project VP9 Encoder\s+(.*)",
            run_cli(f"{get_binary('vpxenc')} --help").get_output(),
        )
        if match is None:
            raise Exception("FATAL: Could not find av1 encoder version")
        return match.group(1)



"""
Provides a central place to get the path to the binaries, and checks if they exist.
Also includes checks if the binaries are what we need (e.g., ffmpeg has been compiled with certain flags)
"""

import os
from shutil import which

__all__ = ["get_binary", "register_bin", "verify_ffmpeg_library", "check_bin"]

from typing import List

from alabamaEncode.core.util.cli_executor import run_cli

bins = []


def check_bin(path) -> bool:
    if path is None:
        return False
    _which = which(path) is not None
    if _which:
        return True
    else:
        if os.path.exists(path):
            return True
        else:
            return False


class FFmpegNotCompiledWithLibrary(Exception):
    def __init__(self, lib_name):
        self.lib_name = lib_name

    def __str__(self):
        return f"ffmpeg is not compiled with {self.lib_name}"


ffmpeg_build_conf = ""


def check_ffmpeg_libraries(lib_name: str) -> bool:
    """
    Checks if the ffmpeg libraries are compiled with the given library
    :param lib_name: name of the library
    :return: True if the library is compiled, False otherwise
    """
    global ffmpeg_build_conf
    if ffmpeg_build_conf == "":
        ffmpeg_build_conf = (
            run_cli(f"{get_binary('ffmpeg')} -v error -buildconf").verify().get_output()
        )
    return ffmpeg_build_conf.find(lib_name) != -1


def verify_ffmpeg_library(lib_name: [str | List[str]]) -> None:
    """
    Checks if the ffmpeg libraries are compiled with the given library, and raises an exception if it is not
    :param lib_name: name of the library
    """
    if isinstance(lib_name, str):
        lib_name = [lib_name]
    for lib in lib_name:
        if not check_ffmpeg_libraries(lib):
            raise FFmpegNotCompiledWithLibrary(lib)


def check_for_ffmpeg_libraries(lib_name: [str | List[str]]) -> bool:
    """
    Checks if the ffmpeg libraries are compiled with the given library, and returns True if it is, False otherwise
    :param lib_name: name of the library(s)
    :return: True if the library is compiled, False otherwise
    """
    if isinstance(lib_name, str):
        lib_name = [lib_name]
    for lib in lib_name:
        if not check_ffmpeg_libraries(lib):
            return False
    return True


class BinaryNotFound(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return (
            f"Binary {self.name} not found,"
            f" set the {self.name.upper()}_CLI_PATH environment variable to the path of the binary."
        )


def register_bin(name, cli):
    bins.append((name, cli))


def get_binary(name):
    _bin = os.getenv(f"{name.upper()}_CLI_PATH", name)
    if _bin == name:
        for _name, _cli in bins:
            if _name == name:
                _bin = _cli
                break
    if _bin is None:
        _bin = os.path.expanduser(f"~/.alabamaEncoder/bin/{name}")
    if check_bin(_bin):
        return _bin
    else:
        raise BinaryNotFound(name)



import os

from matplotlib import pyplot as plt

from alabamaEncode.core.util.bin_utils import register_bin
from alabamaEncode.encoder.impl.X264 import EncoderX264
from alabamaEncode.metrics.comparison_display import ComparisonDisplayResolution
from alabamaEncode.scene.scene_detection import scene_detect
from alabamaEncode.scene.sequence import ChunkSequence

if __name__ == "__main__":
    env = "./convexhull"
    env = os.path.abspath(env)
    if not os.path.exists(env):
        os.mkdir(env)

    scenes: ChunkSequence = scene_detect(
        input_file="/home/kokoniara/ep3_halo_test.mkv",
        max_scene_length=10,
        cache_file_path=f"{env}/scenes_skinny.pt",
    )

    register_bin("SvtAv1EncApp", "/home/kokoniara/.local/opt/SvtAv1EncApp")
    target_vmaf = 90
    ref_dispaly = ComparisonDisplayResolution.FHD
    resoultions = [
        # "3840:-2",
        # "2560:-2",
        "1920:-2",
        "1366:-2",
        "1280:-2",
        "960:-2",
        "854:-2",
        "768:-2",
        "640:-2",
        "480:-2",
    ]

    crfs = range(15, 45, 2)
    pikle = os.path.join(env, "results_X264.pkl")
    import pickle

    results = []
    encs = [
        # EncoderSvtenc(),
        # EncoderVPX("vp9")
        EncoderX264()
    ]
    # for enc in encs:
    #     enc.chunk = scenes[4]
    #     enc.speed = 6
    #     if isinstance(enc, EncoderVPX):
    #         enc.speed = 4
    #     if isinstance(enc, EncoderX264):
    #         enc.speed = 2
    #     enc.passes = 1
    #     enc.threads = 12
    #
    #     for res in resoultions:
    #         for crf in crfs:
    #             filename = f"{res.split(':')[0]}p_{crf}_{enc.__class__.__name__}.{enc.get_chunk_file_extension()}"
    #             enc.output_path = os.path.join(env, filename)
    #             enc.video_filters = f"crop=3840:1920:0:120,scale={res}"
    #             enc.crf = crf
    #             stats = enc.run(
    #                 calculate_vmaf=True,
    #                 vmaf_params=VmafOptions(
    #                     ref=ref_dispaly,
    #                 ),
    #             )
    #             os.remove(enc.output_path)
    #             stats.basename = filename.replace(".ivf", "")
    #             stats.version = crf
    #             print(stats)
    #             results.append(stats)

    # with open(pikle, "wb") as f:
    #     pickle.dump(results, f)

    with open(pikle, "rb") as f:
        results = pickle.load(f)

    plt.figure(figsize=(10, 10))
    plt.title("convex hull of VMAF vs BITRATE")
    plt.xlabel("bitrate")
    # bitrate as log
    plt.xscale("log")
    plt.ylabel("VMAF")
    plt.grid(True)
    for enc in encs:
        # if not isinstance(enc, EncoderVPX):
        #     continue
        for res in resoultions:
            x = []
            y = []
            for stat in results:
                if (
                    stat.basename.startswith(f"{res.split(':')[0]}p")
                    and enc.__class__.__name__ in stat.basename
                ):
                    x.append(stat.bitrate)
                    y.append(stat.vmaf)

            plt.plot(x, y, label=f"{res} {enc.__class__.__name__}")

    plt.legend()
    plt.show()
    #
    # # speed vs vmaf
    # plt.figure(figsize=(10, 10))
    # plt.title("speed vs VMAF")
    # plt.xlabel("time_encoding")
    # plt.ylabel("VMAF")
    # plt.grid(True)
    # for enc in encs:
    #     # if not isinstance(enc, EncoderVPX):
    #     #     continue
    #     for res in resoultions:
    #         x = []
    #         y = []
    #         for stat in results:
    #             if stat.basename.startswith(
    #                 f"{res.split(':')[0]}p"
    #             ) and stat.basename.endswith(enc.__class__.__name__):
    #                 x.append(stat.time_encoding)
    #                 y.append(stat.vmaf)
    #
    #         plt.plot(x, y, label=f"{res} {enc.__class__.__name__}")
    #
    # plt.legend()
    # plt.savefig(os.path.join(env, "speed_vmaf.png"))
    # plt.show()

    vmaf_targets = [45, 55, 62, 68, 81, 87, 90, 93, 95, 97]
    for vmaf_target in vmaf_targets:
        # get res that will reach vmaf_target as close as possible, while having the lowest bitrate
        best_stat = None
        for enc in encs:
            # if not isinstance(enc, EncoderVPX):
            #     continue
            # for res in resoultions:
            #     x = []
            #     y = []
            #     z = []
            #     for stat in results:
            #         if (
            #             stat.basename.startswith(f"{res.split(':')[0]}p")
            #             and enc.__class__.__name__ in stat.basename
            #         ):
            #             x.append(stat.bitrate)
            #             y.append(stat.vmaf)
            #             z.append(stat.version)
            #
            #     for i in range(len(x)):
            #         if y[i] >= vmaf_target:
            #             if lowest_bitrate == 0 or x[i] < lowest_bitrate:
            #                 lowest_bitrate = x[i]
            #                 lowest_bitrate_res = res
            #                 closes_vmaf = y[i]
            #                 accompanying_crf = z[i]
            for stat in results:
                if enc.__class__.__name__ in stat.basename:
                    if best_stat is None:
                        best_stat = stat
                    score = abs(stat.vmaf - vmaf_target)
                    best_score = abs(best_stat.vmaf - vmaf_target)
                    if score < best_score:
                        best_stat = stat

        print(
            f"bitrate {best_stat.bitrate} @ {best_stat.version} crf {best_stat.basename.split('_')[0]} for vmaf {vmaf_target} "
            f"(got {best_stat.vmaf})"
        )



import os.path

from tqdm import tqdm

from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.kv import AlabamaKv
from alabamaEncode.core.util.path import PathAlabama


class WrongFrameCountError(Exception):
    def __init__(self, expected_frame_count: int = -1, actual_frame_count: int = -1):
        # super().__init__("The frame count of the chunk is not equal to the expected")
        super().__init__(
            f"Expected {expected_frame_count} frames, got {actual_frame_count}"
        )


class FfmpegDecodeFailException(Exception):
    def __init__(self):
        super().__init__("Ffmpeg failed to decode the video")


class ChunkObject:
    """
    Ffmpeg based video chunk object
    example:
    If we want to extract a chunk that starts at frame 100 and ends at frame 200:
    obj = ChunkObject(100,200,"infile.mkv")
    f'ffmpeg {obj.get_ss_ffmpeg_command_pair()} outfile.mkv'
    """

    def __init__(
        self,
        first_frame_index=-1,
        last_frame_index=-1,
        path="",
        framerate=-1,
        chunk_index=-1,
        height=-1,
        width=-1,
        complexity=-1.0,
    ):
        self.path = path  # path to the source video file
        self.last_frame_index = last_frame_index
        self.first_frame_index = first_frame_index
        self.framerate = framerate
        self.width = width
        self.height = height
        self.complexity = complexity
        self.length = self.last_frame_index - self.first_frame_index
        self.end_override = -1  # ends the chunk after `end_override` frames if set
        self.chunk_index = chunk_index
        self.chunk_path = ""  # path to the encoded (or not yet) chunk file
        self.chunk_done = False
        self.ideal_bitrate = -1
        self.size_kB = -1

    def __str__(self):
        return f"ChunkObject({self.first_frame_index}, {self.last_frame_index}, {self.path}, {self.framerate})"

    def dict(self) -> dict:
        return {
            "first_frame_index": self.first_frame_index,
            "last_frame_index": self.last_frame_index,
            "path": self.path,
            "framerate": self.framerate,
            "width": self.width,
            "height": self.height,
            "complexity": self.complexity,
            "length": self.length,
            "end_override": self.end_override,
            "chunk_index": self.chunk_index,
            "chunk_path": self.chunk_path,
            "chunk_done": self.chunk_done,
            "ideal_bitrate": self.ideal_bitrate,
            "size_kB": self.size_kB,
        }

    def to_json(self) -> str:
        import json

        return json.dumps(self.dict())

    @staticmethod
    def from_json(json_str: str):
        import json

        c = ChunkObject()
        c.__dict__ = json.loads(json_str)
        return c

    def get_frame_count(self) -> int:
        return self.last_frame_index - self.first_frame_index

    def get_lenght(self) -> float:
        # get framerate
        if self.framerate == -1:
            self.framerate = Ffmpeg.get_video_frame_rate(PathAlabama(self.path))

        local_overriden_end = self.last_frame_index
        if self.end_override != -1 and self.length > self.end_override:
            local_overriden_end = self.first_frame_index + self.end_override

        end_thingy = float(local_overriden_end) / self.framerate
        start_time = float(self.first_frame_index) / self.framerate
        return end_thingy - start_time

    def get_filesize(self) -> float:
        if self.size_kB != -1:
            return self.size_kB * 1000

        if os.path.exists(self.chunk_path):
            return os.path.getsize(self.chunk_path)

    def get_ss_ffmpeg_command_pair(self) -> str:
        """
        :return: an '-ss 12 clip.mp4 -t 2' ffmpeg command from start frame index and end frame index
        """

        # case where we don't have a start or end frame index so include the whole video
        if self.first_frame_index == -1 or self.last_frame_index == -1:
            return f' -i "{self.path}" '

        # get framerate
        if self.framerate == -1:
            self.framerate = Ffmpeg.get_video_frame_rate(PathAlabama(self.path))

        # if we override the end, we end at "start frame # + override"
        local_overriden_end = self.last_frame_index
        if self.end_override != -1 and self.length > self.end_override:
            local_overriden_end = self.first_frame_index + self.end_override

        # get the start time and duration
        end_thingy = float(local_overriden_end) / self.framerate
        start_time = float(self.first_frame_index) / self.framerate
        duration = end_thingy - start_time

        return f' -ss {str(start_time)} -i "{self.path}" -t {str(duration)} '

    def get_width(self) -> int:
        if self.width == -1:
            self.width = Ffmpeg.get_width(PathAlabama(self.path))
        return self.width

    def get_height(self) -> int:
        if self.height == -1:
            self.height = Ffmpeg.get_height(PathAlabama(self.path))
        return self.height

    def create_chunk_ffmpeg_pipe_command(self, video_filters="", bit_depth=10) -> str:
        """
        :param video_filters: ffmpeg vf filters, e.g., scaling tonemapping
        :param bit_depth: bit depth of the output stream 8 or 10
        :return: a 'ffmpeg ... |' command string that pipes a y4m stream into stdout
        """
        if video_filters is None:
            video_filters = ""
        end_command = (
            f"ffmpeg -threads 1 -v error -nostdin -hwaccel auto {self.get_ss_ffmpeg_command_pair()} "
            f"-pix_fmt yuv420p10le "
        )

        if bit_depth == 8:
            end_command = end_command.replace("10le", "")

        if "-vf" not in video_filters and not video_filters == "":
            video_filters = f"-vf {video_filters}"

        end_command += f" -an -sn -strict -1 {video_filters} -f yuv4mpegpipe - "

        return end_command

    def log_prefix(self):
        return f"[{self.chunk_index}] "

    def verify_integrity(self, length_of_sequence=-1, quiet=False) -> bool:
        """
        checks the integrity of a chunk
        :return: True if invalid
        """
        self.chunk_done = False

        if not os.path.exists(self.chunk_path):
            return True

        try:
            path = PathAlabama(self.chunk_path)
            if Ffmpeg.check_for_invalid(path):
                raise FfmpegDecodeFailException()

            actual_frame_count = Ffmpeg.get_frame_count(path)
            expected_frame_count = self.last_frame_index - self.first_frame_index

            if actual_frame_count != expected_frame_count:
                if (
                    length_of_sequence != -1
                    and length_of_sequence == self.chunk_index + 1
                ):
                    if not quiet:
                        print(
                            f"{self.log_prefix()}Frame count mismatch, but it's the last chunk, so it's ok"
                        )
                else:
                    raise WrongFrameCountError(
                        actual_frame_count=actual_frame_count,
                        expected_frame_count=expected_frame_count,
                    )
        except Exception as e:
            if isinstance(e, WrongFrameCountError) or isinstance(
                e, FfmpegDecodeFailException
            ):
                if not quiet:
                    tqdm.write(
                        f"{self.log_prefix()} failed the integrity because: {e} 🤕"
                    )
            return True
        self.size_kB = self.get_filesize() / 1000
        self.chunk_done = True
        return False

    def is_done(self, quiet=False, kv: AlabamaKv = None, length_of_sequence=-1) -> bool:
        """
        checks if the chunk is done
        :param quiet log what's wrong with the chunk to stdout
        :param kv used to cache the integrity calculation
        :param length_of_sequence pass to integ check
        :return: True if done
        """

        if self.chunk_done is True:
            return self.chunk_done

        if kv:
            valid = kv.get("chunk_integrity", self.chunk_index)

            # do an additional check if the chunk file exists
            valid = valid and os.path.exists(self.chunk_path)

            if valid is not None and valid is True:
                self.chunk_done = True
                # print(f"Chunk {chunk.chunk_index} is valid from cache")
                return self.chunk_done

        self.verify_integrity(quiet=quiet, length_of_sequence=length_of_sequence)
        return self.chunk_done


def test_1():
    chunk = ChunkObject(
        path="video.mkv", first_frame_index=100, last_frame_index=200, framerate=24
    )
    expected_1 = " -ss 4.166666666666667 -i video.mkv -t 4.166666666666667 "
    result_1 = chunk.get_ss_ffmpeg_command_pair()
    print(result_1)
    if expected_1 == result_1:
        print("Test 1 passed")

    expected_2 = " -ss 4.166666666666667 -i video.mkv -t 4.166666666666667 "
    chunk.end_override = 150
    result_2 = chunk.get_ss_ffmpeg_command_pair()
    print(result_2)
    if result_2 == expected_2:
        print("Test 2 passed")

    expected_3 = " -ss 4.166666666666667 -i video.mkv -t 2.083333333333333 "
    chunk.end_override = 50
    result_3 = chunk.get_ss_ffmpeg_command_pair()
    print(result_3)
    if result_3 == expected_3:
        print("Test 3 passed")


def test_2():
    chunk = ChunkObject(path="video.mkv")
    expected_1 = " -i video.mkv "
    result_1 = chunk.get_ss_ffmpeg_command_pair()
    print(result_1)
    if expected_1 == result_1:
        print("Test 1 passed")



import json
import os
import threading


class AlabamaKv(object):
    """
    AlabamaKv is a key-value store that stores data in a bucket, each bucket is file in a folder. Stored in json format.
    It is used to store data between runs of alabamaEncode. Saving to files immediately because making this
    thread safe is not worth it/needed.
    Prototypes:
    __init__(folder for buckets)
    set(bucket, key, value)
    get(bucket, key) -> str
    get_all(bucket) -> dict
    exists(bucket, key) -> bool
    get_global(key) -> str  # shortcut for get("kv", key)
    set_global(key, value)  # shortcut for set("kv", key, value)
    """

    def __init__(self, folder):
        self.mutex = threading.Lock()
        self.folder = folder
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)

    def get_global(self, key):
        return self.get("kv", key)

    def set_global(self, key, value):
        return self.set("kv", key, value)

    def set(self, bucket, key, value, individual_mode=False):
        with self.mutex:
            if individual_mode:
                bucket_path = os.path.join(self.folder, bucket)
                if not os.path.exists(bucket_path):
                    os.makedirs(bucket_path)
                key_path = os.path.join(bucket_path, f"{key}.json")
                with open(key_path, "w") as f:
                    json.dump(value, f)
            else:
                bucket_content = self._load(bucket)
                bucket_content[key] = value
                bucket_path = os.path.join(self.folder, bucket + ".json")
                with open(bucket_path, "w") as f:
                    json.dump(bucket_content, f)

    def get(self, bucket, key) -> [str | None]:
        with self.mutex:
            b = self._load(bucket)
            if not isinstance(key, str):
                key = str(key)
            if key not in b:
                return None
            return b[key]

    def get_all(self, bucket):
        with self.mutex:
            b = self._load(bucket)
            return b

    def exists(self, bucket, key):
        with self.mutex:
            b = self._load(bucket)
            return key in b

    def _load(self, bucket_name: str) -> dict:
        bucket_path = os.path.join(self.folder, bucket_name)
        single_file_path = bucket_path + ".json"
        if os.path.exists(single_file_path):
            with open(single_file_path, "r") as f:
                return json.load(f)
        elif os.path.exists(bucket_path):
            bucket_content = {}
            for key_file in os.listdir(bucket_path):
                key = os.path.splitext(key_file)[0]
                with open(os.path.join(bucket_path, key_file)) as f:
                    bucket_content[key] = json.load(f)
            return bucket_content
        else:
            return {}



import copy
import os
import random
import time
from multiprocessing.pool import ThreadPool
from statistics import mean
from typing import List

from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.util.kv import AlabamaKv
from alabamaEncode.encoder.impl.SvtAvif import AvifEncoderSvtenc
from alabamaEncode.metrics.image import ImageMetrics
from alabamaEncode.scene.sequence import ChunkSequence


def setup_autograin(ctx: AlabamaContext, sequence: ChunkSequence):
    if (
        ctx.prototype_encoder.grain_synth == -1
        and ctx.prototype_encoder.supports_grain_synth()
    ):
        ctx.prototype_encoder.grain_synth = get_best_avg_grainsynth(
            input_file=sequence.input_file,
            scenes=sequence,
            temp_folder=ctx.temp_folder,
            video_filters=ctx.prototype_encoder.video_filters,
            kv=ctx.get_kv(),
            crf=ctx.prototype_encoder.crf,
        )

    return ctx


def find_lowest_x(x_list: List[float], y_list: List[float]) -> float:
    # Check that x_list and y_list are the same length
    if len(x_list) != len(y_list):
        raise ValueError("x_list and y_list must have the same length")

    # Find the minimum y-value and its index
    min_y, min_y_idx = min((y, idx) for idx, y in enumerate(y_list))

    # If the minimum y-value is at the beginning or end of the list, return the corresponding x-value
    if min_y_idx == 0:
        return x_list[0]
    elif min_y_idx == len(y_list) - 1:
        return x_list[-1]

    # Otherwise, use linear interpolation to find the x-value that corresponds to the minimum y-value
    x0 = x_list[min_y_idx - 1]
    x1 = x_list[min_y_idx]
    y0 = y_list[min_y_idx - 1]
    y1 = y_list[min_y_idx]
    slope = (y1 - y0) / (x1 - x0)
    x_min = (min_y - y0) / slope + x0

    return x_min


def get_ideal_grain_butteraugli(encoded_scene_path, chunk, crf, bitrate, vf) -> int:
    start = time.time()

    if "vf" not in vf and vf != "":
        vf = f"-vf {vf}"

    avif_enc = AvifEncoderSvtenc()

    if bitrate != -1:
        avif_enc.bitrate = bitrate
    else:
        avif_enc.crf = crf

    # Create a reference png
    ref_png = encoded_scene_path + ".png"
    if not os.path.exists(ref_png):
        cmd = (
            f'{get_binary("ffmpeg")} -hide_banner -y {chunk.get_ss_ffmpeg_command_pair()} '
            f'{vf} -frames:v 1 "{ref_png}"'
        )

        out = run_cli(cmd)
        if not os.path.exists(ref_png):
            print(cmd)
            raise Exception(f"Could not create reference png: {out}")

    avif_enc.in_path = ref_png

    runs = []

    grain_probes = [0, 1, 4, 6, 11, 16, 21, 26]
    for grain in grain_probes:
        avif_enc.grain_synth = grain
        avif_enc.output_path = encoded_scene_path + ".grain" + str(grain) + ".avif"
        avif_enc.run()

        if not os.path.exists(avif_enc.output_path):
            raise Exception("Encoding of avif Failed")

        decoded_test_png_path = avif_enc.output_path + ".png"

        # turn the avif into a png
        run_cli(
            f'{get_binary("ffmpeg")} -y -i "{avif_enc.output_path}" "{decoded_test_png_path}"'
        )

        if not os.path.exists(decoded_test_png_path):
            raise Exception("Could not create decoded png")

        rd = {
            "grain": grain,
            "butter": ImageMetrics.butteraugli_score(ref_png, decoded_test_png_path),
        }

        os.remove(decoded_test_png_path)
        os.remove(avif_enc.output_path)

        print(f"{chunk.log_prefix()} grain {rd['grain']} -> {rd['butter']} butteraugli")
        runs.append(rd)

    os.remove(ref_png)

    # find the film-grain value that corresponds to the lowest butteraugli score
    ideal_grain = find_lowest_x(
        [point["grain"] for point in runs], [point["butter"] for point in runs]
    )

    print(f"ideal grain is {ideal_grain}, in {int(time.time() - start)} seconds")
    return int(ideal_grain)


def wrapper(obj):
    return get_ideal_grain_butteraugli(**obj)


def get_best_avg_grainsynth(
    input_file: str,
    scenes: ChunkSequence,
    temp_folder="./grain_test",
    kv: AlabamaKv = None,
    bitrate=-1,
    crf=20,
    video_filters: str = "",
) -> int:
    cache_key = "best_sequence_grain"

    if kv is not None:
        val = kv.get_global(cache_key)
        if val is not None:
            return int(val)

    if not os.path.exists(temp_folder):
        raise Exception(f"temp_folder {temp_folder} does not exist")

    # turn temp folder into a full path
    temp_folder = os.path.abspath(temp_folder)
    # make /adapt/grain dir
    os.makedirs(f"{temp_folder}/adapt/grain", exist_ok=True)

    if input_file is None:
        raise Exception("input_file is required")

    if scenes is None:
        raise Exception("scenes is required")
    # create a copy of the object, so it doesn't cause trouble
    scenes = copy.deepcopy(scenes)

    print("starting autograin test")

    if len(scenes.chunks) > 10:
        # bases on length, remove every x scene from the list so its shorter
        scenes.chunks = scenes.chunks[:: int(len(scenes.chunks) / 10)]

    random.seed(2)
    random.shuffle(scenes.chunks)

    chunks_for_processing = scenes.chunks[:6]

    jobs = [
        {
            "chunk": chunk,
            "encoded_scene_path": f"{temp_folder}/adapt/grain/{chunk.chunk_index}",
            "crf": crf,
            "bitrate": bitrate,
            "vf": video_filters,
        }
        for chunk in chunks_for_processing
    ]

    # parallelize the butteraugli tests
    with ThreadPool() as p:
        results = p.map(wrapper, jobs)
        p.close()
        p.join()

    # get the results
    ideal_grain = int(mean(results))
    print(f"for 6 random scenes, the average ideal grain is {ideal_grain}")
    if kv is not None:
        kv.set_global(cache_key, ideal_grain)
    return ideal_grain



import json
import os
import re
import time
from typing import Any

from alabamaEncode.core.util.bin_utils import get_binary, verify_ffmpeg_library
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.util.path import PathAlabama


class Ffmpeg:
    @staticmethod
    def check_for_invalid(path: PathAlabama) -> bool:
        """
        Checks if the file is not a valid video file
        :param path: PathAlabama instance
        :return: True if the video is invalid, False if the video is valid
        """
        path.check_video()

        return not run_cli(
            f"ffmpeg -v error -i {path.get_safe()} -c copy -f null /dev/null"
        ).success()

    @staticmethod
    def get_frame_count(path: PathAlabama) -> int:
        """
        Returns the frame count of the video
        :param path: path to the video
        :return: int
        """
        path.check_video()
        return (
            run_cli(
                f"ffprobe -v error -select_streams v:0 -count_packets -show_entries stream=nb_read_packets "
                f"-of default=noprint_wrappers=1:nokey=1 {path.get_safe()}"
            )
            .verify()
            .strip_mp4_warning()
            .get_as_int()
        )

    @staticmethod
    def get_frame_count_fast(path: PathAlabama):
        length_in_secs = Ffmpeg.get_video_length(path)
        fps = Ffmpeg.get_video_frame_rate(path)
        return int(length_in_secs * fps)

    @staticmethod
    def get_tracks(path: PathAlabama):
        path.check_video()
        out = (
            run_cli(
                f'{get_binary("ffprobe")} -v error -show_entries stream -of json {path.get_safe()}'
            )
            .verify()
            .strip_mp4_warning()
            .get_output()
        )
        return json.loads(out)["streams"]

    @staticmethod
    def get_video_length(path: PathAlabama, sexagesimal=False) -> float | str:
        """
        Returns the video length in seconds
        :param sexagesimal: If True, returns the length in sexagesimal format
        :param path: Path to the video
        :return: float
        """
        path.check_video()
        cli_command = (
            f"ffprobe -v error -show_entries format=duration {'-sexagesimal' if sexagesimal else ''}"
            f" -of default=noprint_wrappers=1:nokey=1 {path.get_safe()}"
        )
        try:
            out = (
                run_cli(cli_command)
                .verify(
                    bad_output_hints=["N/A", "Invalid data found"],
                    fail_message=f"ffprobe failed, {cli_command}",
                )
                .strip_mp4_warning()
                .get_output()
            )
        except RuntimeError:
            frame_count = Ffmpeg.get_frame_count(path)
            fps = Ffmpeg.get_video_frame_rate(path)
            return frame_count / fps

        return out if sexagesimal else float(out)

    @staticmethod
    def get_total_bitrate(path: PathAlabama) -> float:
        path.check_video()
        return (os.path.getsize(path.get()) * 8) / Ffmpeg.get_video_length(path)

    @staticmethod
    def get_height(path: PathAlabama) -> int:
        path.check_video()
        return (
            run_cli(
                f"ffprobe -v error -select_streams v:0 -show_entries stream=height "
                f"-of default=noprint_wrappers=1:nokey=1 {path.get_safe()}"
            )
            .verify()
            .strip_mp4_warning()
            .get_as_int()
        )

    @staticmethod
    def get_width(path: PathAlabama) -> int:
        path.check_video()
        return (
            run_cli(
                f"ffprobe -v error -select_streams v:0 -show_entries stream=width "
                f"-of default=noprint_wrappers=1:nokey=1 {path.get_safe()}"
            )
            .verify()
            .strip_mp4_warning()
            .get_as_int()
        )

    @staticmethod
    def get_pix_fmt(path: PathAlabama) -> str:
        path.check_video()
        return (
            run_cli(
                f"ffprobe -v error -select_streams v:0 -show_entries stream=pix_fmt "
                f"-of default=noprint_wrappers=1:nokey=1 {path.get_safe()}"
            )
            .verify()
            .strip_mp4_warning()
            .get_output()
        )

    @staticmethod
    def get_bit_depth(path: PathAlabama) -> int:
        pix_fmt = Ffmpeg.get_pix_fmt(path)
        if pix_fmt == "yuv420p":
            return 8
        elif pix_fmt == "yuv420p10le":
            return 10
        else:
            raise NotImplemented(
                f"could not parse bitdepth out of: {pix_fmt}; please report it"
            )

    @staticmethod
    def is_hdr(path: PathAlabama) -> bool:
        """Check if a video is HDR"""
        path.check_video()
        out = (
            run_cli(
                (
                    f"ffprobe -v quiet -show_entries stream=color_transfer "
                    f"-of csv=p=0 -select_streams v:0 {path.get_safe()}"
                )
            )
            .get_output()
            .strip()
        )

        if "bt709" in out or "unknown" in out:
            return False
        return True

    @staticmethod
    def get_video_frame_rate(file: PathAlabama) -> float:
        file.check_video()
        result = (
            run_cli(
                (
                    f"ffprobe -v error -select_streams v -of default=noprint_wrappers=1:nokey=1"
                    f" -show_entries stream=r_frame_rate {file.get_safe()}"
                )
            )
            .verify(bad_output_hints=[""])
            .strip_mp4_warning()
            .get_output()
            .split("/")
        )
        return float(result[0]) / float(result[1])

    @staticmethod
    def get_fps_fraction(file: PathAlabama) -> str:
        file.check_video()
        return (
            run_cli(
                (
                    f"ffprobe -v error -select_streams v -of default=noprint_wrappers=1:nokey=1"
                    f" -show_entries stream=r_frame_rate {file.get_safe()}"
                )
            )
            .verify(bad_output_hints=[""])
            .strip_mp4_warning()
            .get_output()
        )

    @staticmethod
    def get_source_bitrates(
        path: PathAlabama, calculate_video=True, calculate_audio=True
    ) -> tuple[float, float]:
        """
        stolen from the one and only autocompressor.com's source code 🤑
        Returns tuple of bitrates (firstVideoStream, firstAudioStream)
        Works via demux-to-null (container stats are considered false)
        """
        common = "-show_entries packet=size -of default=nokey=1:noprint_wrappers=1"
        command_v = f"ffprobe -v error -select_streams V:0 {common} {path.get_safe()}"
        command_a = f"ffprobe -v error -select_streams a:0 {common} {path.get_safe()}"
        vid_bps = 0
        aud_bps = 0

        real_duration = Ffmpeg.get_video_length(path)

        if calculate_video:
            packets_v_arr = (
                run_cli(command_v).verify().strip_mp4_warning().get_output().split("\n")
            )
            packets_v_bits = sum([int(i) * 8 for i in packets_v_arr if i.isdigit()])
            vid_bps = round(packets_v_bits / real_duration)

        if calculate_audio:
            packets_a_arr = (
                run_cli(command_a).verify().strip_mp4_warning().get_output().split("\n")
            )
            packets_a_bits = sum([int(j) * 8 for j in packets_a_arr if j.isdigit()])
            aud_bps = round(packets_a_bits / real_duration)

        return vid_bps, aud_bps

    @staticmethod
    def get_tonemap_vf() -> str:
        # https://ffmpeg.org/ffmpeg-filters.html#tonemap
        verify_ffmpeg_library("libzimg")
        return "zscale=t=linear,tonemap=mobius,zscale=p=bt709:t=bt709:m=bt709:r=tv:d=error_diffusion"

    @staticmethod
    def get_frame_data(path: PathAlabama) -> dict:
        path.check_video()

        return json.loads(
            run_cli(
                f"ffprobe -v error -select_streams v:0 -show_frames -of json {path.get_safe()}"
            )
            .verify()
            .get_output()
        )

    @staticmethod
    def get_first_frame_data(path: PathAlabama) -> dict:
        # ffprobe -v error -select_streams v:0 -show_frames 'path' -of json -read_intervals "%+0.3"
        path.check_video()
        return json.loads(
            run_cli(
                f'ffprobe -v error -select_streams v:0 -show_frames -of json -read_intervals "1%+#1" {path.get_safe()}'
            )
            .verify()
            .get_output()
        )["frames"][0]

    @staticmethod
    def get_codec(path: PathAlabama) -> str:
        path.check_video()
        return (
            run_cli(
                f"ffprobe -v error -select_streams v:0 -show_entries stream=codec_name "
                f"-of default=noprint_wrappers=1:nokey=1 {path.get_safe()}"
            )
            .verify()
            .strip_mp4_warning()
            .get_output()
        )

    @staticmethod
    def get_vmaf_motion(chunk) -> float:
        verify_ffmpeg_library("libvmaf")
        # [Parsed_vmafmotion_0 @ 0x558626bfa300] VMAF Motion avg: 8.732
        out = (
            run_cli(
                f"{get_binary('ffmpeg')} {chunk.get_ss_ffmpeg_command_pair()} -lavfi vmafmotion -f null /dev/null "
            )
            .verify()
            .strip_mp4_warning()
            .get_output()
        )
        return float(re.findall(r"VMAF Motion avg: ([0-9.]+)", out)[0])

    @staticmethod
    def get_ffprobe_content_features(chunk, vf) -> dict[Any, Any]:
        # ffmpeg -v error -nostdin -hwaccel auto -i
        # '/home/kokoniara/howlscastle_test.mkv'  -pix_fmt yuv420p10le -an -sn -strict -1 -f yuv4mpegpipe - | ffprobe
        # -v error -select_streams v:0 -f lavfi -i 'movie=/dev/stdin,entropy,scdet,signalstats' -show_frames -of json

        out = (
            run_cli(
                f"{chunk.create_chunk_ffmpeg_pipe_command(video_filters=vf)} | {get_binary('ffprobe')} -v error "
                f"-select_streams v:0 -f lavfi -i 'movie=/dev/stdin,entropy,scdet,signalstats' -show_frames -of json"
            )
            .verify()
            .get_output()
        )

        parsed_json = json.loads(out)
        filtered = {}

        for frame in parsed_json["frames"]:
            tags = frame["tags"]

            tags = {
                "entropy.entropy.normal.Y": tags["lavfi.entropy.entropy.normal.Y"],
                "entropy.normalized_entropy.normal.Y": tags[
                    "lavfi.entropy.normalized_entropy.normal.Y"
                ],
                "entropy.entropy.normal.U": tags["lavfi.entropy.entropy.normal.U"],
                "entropy.normalized_entropy.normal.U": tags[
                    "lavfi.entropy.normalized_entropy.normal.U"
                ],
                "entropy.entropy.normal.V": tags["lavfi.entropy.entropy.normal.V"],
                "entropy.normalized_entropy.normal.V": tags[
                    "lavfi.entropy.normalized_entropy.normal.V"
                ],
                "scd.mafd": tags["lavfi.scd.mafd"],
                "scd.score": tags["lavfi.scd.score"],
                "signalstats.YMIN": tags["lavfi.signalstats.YMIN"],
                "signalstats.YLOW": tags["lavfi.signalstats.YLOW"],
                "signalstats.YAVG": tags["lavfi.signalstats.YAVG"],
                "signalstats.YHIGH": tags["lavfi.signalstats.YHIGH"],
                "signalstats.YMAX": tags["lavfi.signalstats.YMAX"],
                "signalstats.UMIN": tags["lavfi.signalstats.UMIN"],
                "signalstats.ULOW": tags["lavfi.signalstats.ULOW"],
                "signalstats.UAVG": tags["lavfi.signalstats.UAVG"],
                "signalstats.UHIGH": tags["lavfi.signalstats.UHIGH"],
                "signalstats.UMAX": tags["lavfi.signalstats.UMAX"],
                "signalstats.VMIN": tags["lavfi.signalstats.VMIN"],
                "signalstats.VLOW": tags["lavfi.signalstats.VLOW"],
                "signalstats.VAVG": tags["lavfi.signalstats.VAVG"],
                "signalstats.VHIGH": tags["lavfi.signalstats.VHIGH"],
                "signalstats.VMAX": tags["lavfi.signalstats.VMAX"],
                "signalstats.SATMIN": tags["lavfi.signalstats.SATMIN"],
                "signalstats.SATLOW": tags["lavfi.signalstats.SATLOW"],
                "signalstats.SATAVG": tags["lavfi.signalstats.SATAVG"],
            }

            filtered[frame["pts"]] = tags
        return filtered

    @staticmethod
    def get_siti_tools_data(chunk, vf) -> dict[Any, Any]:
        out = (
            run_cli(
                f"{chunk.create_chunk_ffmpeg_pipe_command(video_filters=vf)} |"
                f" {get_binary('siti-tools')} -f json -b 10 -r full -q /dev/stdin"
            )
            .verify()
            .get_output()
        )

        parsed_json = json.loads(out)
        # remove "settings" and "input_file" keys
        parsed_json.pop("settings", None)
        parsed_json.pop("input_file", None)

        # {
        #     "si": [
        #       1,
        #       1,
        #       for every frame...
        #     ],
        #      "ti": [
        #       1,
        #       1,
        #       for every frame except the first...
        #     ]
        # }
        # convert to:
        # {
        #     "frame #": [
        #         si score
        #         ti score # 0 if first frame
        #     ]
        #   ... for every frame
        # }
        new = []
        for i in range(len(parsed_json["si"])):
            si_ti_pair = [parsed_json["si"][i], 0]
            if i != 0:
                si_ti_pair[1] = parsed_json["ti"][i - 1]
            new.append(si_ti_pair)

        return_dict = {}
        for i, frame in enumerate(new):
            return_dict[i] = frame

        return return_dict

    @staticmethod
    def get_content_features(chunk, vf=""):
        siti = Ffmpeg.get_siti_tools_data(chunk, vf)
        ffprobe = Ffmpeg.get_ffprobe_content_features(chunk, vf)
        # since ffprobe and siti are both dicts that use frame # as the key, we can just merge them
        # and return the result
        new = {}
        for key in ffprobe.keys():
            new[key] = {
                **ffprobe[key],
                "siti.si": siti[key][0],
                "siti.ti": siti[key][1],
            }
        return new


def track_test():
    tracks = Ffmpeg.get_tracks(PathAlabama("/home/kokoniara/owoStreamCopy_ffv1.mkv"))
    for track in tracks:
        match track["codec_type"]:
            case "video":
                print("video track")
            case "audio":
                print("audio track")
            case "subtitle":
                print("subtitle track")
        print(track)


def test_frame_data():
    cache_file = "owo_cache"
    print("Scraping dv meta using ffprobe...")
    start = time.time()
    if not os.path.exists(cache_file):
        frame_data = Ffmpeg.get_frame_data(
            PathAlabama("/home/kokoniara/ep3_halo_test.mkv")
        )
        with open(cache_file, "w") as f:
            f.write(json.dumps(frame_data))
    else:
        with open(cache_file) as f:
            frame_data = json.loads(f.read())

    # print(frame_data)
    print(f"Finished in {time.time() - start}s")

    for i, frame in enumerate(frame_data["frames"]):
        # print("\n\n")
        # print(frame["side_data_list"][3]["scene_refresh_flag"])
        new_scene = frame["side_data_list"][3]["scene_refresh_flag"]
        if new_scene == 1:
            print(f"{i} is a new scene")


if __name__ == "__main__":
    # track_test()
    test_frame_data()
